import glob
import operator
import os
import signal
import warnings
from fnmatch import fnmatch
try:
    import resource
except ImportError:
    resource = None     # NOQA

from circus import logger
from circus.util import (DEFAULT_ENDPOINT_DEALER, DEFAULT_ENDPOINT_SUB,
                         DEFAULT_ENDPOINT_MULTICAST, DEFAULT_ENDPOINT_STATS,
                         StrictConfigParser, replace_gnu_args, to_signum,
                         to_bool)


def watcher_defaults():
    return {
        'name': '',
        'cmd': '',
        'args': '',
        'numprocesses': 1,
        'warmup_delay': 0,
        'executable': None,
        'working_dir': None,
        'on_demand': False,
        'shell': False,
        'uid': None,
        'gid': None,
        'send_hup': False,
        'stop_signal': signal.SIGTERM,
        'stop_children': False,
        'max_retry': 5,
        'graceful_timeout': 30,
        'rlimits': dict(),
        'stderr_stream': dict(),
        'stdout_stream': dict(),
        'priority': 0,
        'use_sockets': False,
        'singleton': False,
        'copy_env': False,
        'copy_path': False,
        'hooks': dict(),
        'respawn': True,
        'autostart': True}


class DefaultConfigParser(StrictConfigParser):

    def __init__(self, *args, **kw):
        StrictConfigParser.__init__(self, *args, **kw)
        self._env = dict(os.environ)

    def set_env(self, env):
        self._env = dict(env)

    def get(self, section, option, **kwargs):
        res = StrictConfigParser.get(self, section, option, **kwargs)
        return replace_gnu_args(res, env=self._env)

    def items(self, section, noreplace=False):
        items = StrictConfigParser.items(self, section)
        if noreplace:
            return items

        return [(key, replace_gnu_args(value, env=self._env))
                for key, value in items]

    def dget(self, section, option, default=None, type=str):
        if not self.has_option(section, option):
            return default

        value = self.get(section, option)

        if type is int:
            value = int(value)
        elif type is bool:
            value = to_bool(value)
        elif type is float:
            value = float(value)
        elif type is not str:
            raise NotImplementedError()

        return value


def rlimit_value(val):
    if resource is not None and (val is None or len(val) == 0):
        return resource.RLIM_INFINITY
    else:
        return int(val)


def read_config(config_path):
    cfg = DefaultConfigParser()
    with open(config_path) as f:
        if hasattr(cfg, 'read_file'):
            cfg.read_file(f)
        else:
            cfg.readfp(f)

    current_dir = os.path.dirname(config_path)

    # load included config files
    includes = []

    def _scan(filename, includes):
        if os.path.abspath(filename) != filename:
            filename = os.path.join(current_dir, filename)

        paths = glob.glob(filename)
        if paths == []:
            logger.warn('%r does not lead to any config. Make sure '
                        'include paths are relative to the main config '
                        'file' % filename)
        includes += paths

    for include_file in cfg.dget('circus', 'include', '').split():
        _scan(include_file, includes)

    for include_dir in cfg.dget('circus', 'include_dir', '').split():
        _scan(os.path.join(include_dir, '*.ini'), includes)

    logger.debug('Reading config files: %s' % includes)
    return cfg, [config_path] + cfg.read(includes)


def get_config(config_file):
    if not os.path.exists(config_file):
        raise IOError("the configuration file %r does not exist\n" %
                      config_file)

    cfg, cfg_files_read = read_config(config_file)
    dget = cfg.dget
    config = {}

    # reading the global environ first
    global_env = dict(os.environ.items())
    local_env = dict()

    # update environments with [env] section
    if 'env' in cfg.sections():
        local_env.update(dict(cfg.items('env')))
        global_env.update(local_env)

    # always set the cfg environment
    cfg.set_env(global_env)

    # main circus options
    config['check_delay'] = dget('circus', 'check_delay', 5., float)
    config['endpoint'] = dget('circus', 'endpoint', DEFAULT_ENDPOINT_DEALER)
    config['endpoint_owner'] = dget('circus', 'endpoint_owner', None, str)
    config['pubsub_endpoint'] = dget('circus', 'pubsub_endpoint',
                                     DEFAULT_ENDPOINT_SUB)
    config['multicast_endpoint'] = dget('circus', 'multicast_endpoint',
                                        DEFAULT_ENDPOINT_MULTICAST)
    config['stats_endpoint'] = dget('circus', 'stats_endpoint', None)
    config['statsd'] = dget('circus', 'statsd', False, bool)
    config['umask'] = dget('circus', 'umask', None)
    if config['umask']:
        config['umask'] = int(config['umask'], 8)

    if config['stats_endpoint'] is None:
        config['stats_endpoint'] = DEFAULT_ENDPOINT_STATS
    elif not config['statsd']:
        warnings.warn("You defined a stats_endpoint without "
                      "setting up statsd to True.",
                      DeprecationWarning)
        config['statsd'] = True

    config['warmup_delay'] = dget('circus', 'warmup_delay', 0, int)
    config['httpd'] = dget('circus', 'httpd', False, bool)
    config['httpd_host'] = dget('circus', 'httpd_host', 'localhost', str)
    config['httpd_port'] = dget('circus', 'httpd_port', 8080, int)
    config['debug'] = dget('circus', 'debug', False, bool)
    config['debug_gc'] = dget('circus', 'debug_gc', False, bool)
    config['pidfile'] = dget('circus', 'pidfile')
    config['loglevel'] = dget('circus', 'loglevel')
    config['logoutput'] = dget('circus', 'logoutput')
    config['loggerconfig'] = dget('circus', 'loggerconfig', None)
    config['fqdn_prefix'] = dget('circus', 'fqdn_prefix', None, str)

    # Initialize watchers, plugins & sockets to manage
    watchers = []
    plugins = []
    sockets = []

    for section in cfg.sections():
        section_items = dict(cfg.items(section))
        if list(section_items.keys()) in [[], ['__name__']]:
            # Skip empty sections
            continue
        if section.startswith("socket:"):
            sock = section_items
            sock['name'] = section.split("socket:")[-1].lower()
            sock['so_reuseport'] = dget(section, "so_reuseport", False, bool)
            sock['replace'] = dget(section, "replace", False, bool)
            sockets.append(sock)

        if section.startswith("plugin:"):
            plugin = section_items
            plugin['name'] = section
            if 'priority' in plugin:
                plugin['priority'] = int(plugin['priority'])
            plugins.append(plugin)

        if section.startswith("watcher:"):
            watcher = watcher_defaults()
            watcher['name'] = section.split("watcher:", 1)[1]

            # create watcher options
            for opt, val in cfg.items(section, noreplace=True):
                if opt in ('cmd', 'args', 'working_dir', 'uid', 'gid'):
                    watcher[opt] = val
                elif opt == 'numprocesses':
                    watcher['numprocesses'] = dget(section, 'numprocesses', 1,
                                                   int)
                elif opt == 'warmup_delay':
                    watcher['warmup_delay'] = dget(section, 'warmup_delay', 0,
                                                   int)
                elif opt == 'executable':
                    watcher['executable'] = dget(section, 'executable', None,
                                                 str)
                # default bool to False
                elif opt in ('on_demand', 'shell', 'send_hup', 'stop_children',
                             'close_child_stderr', 'use_sockets', 'singleton',
                             'copy_env', 'copy_path', 'close_child_stdout'):
                    watcher[opt] = dget(section, opt, False, bool)
                elif opt == 'stop_signal':
                    watcher['stop_signal'] = to_signum(val)
                elif opt == 'max_retry':
                    watcher['max_retry'] = dget(section, "max_retry", 5, int)
                elif opt == 'graceful_timeout':
                    watcher['graceful_timeout'] = dget(
                        section, "graceful_timeout", 30., float)
                elif opt.startswith('stderr_stream') or \
                        opt.startswith('stdout_stream'):
                    stream_name, stream_opt = opt.split(".", 1)
                    watcher[stream_name][stream_opt] = val
                elif opt.startswith('rlimit_'):
                    limit = opt[7:]
                    watcher['rlimits'][limit] = rlimit_value(val)
                elif opt == 'priority':
                    watcher['priority'] = dget(section, "priority", 0, int)
                elif opt.startswith('hooks.'):
                    hook_name = opt[len('hooks.'):]
                    val = [elmt.strip() for elmt in val.split(',', 1)]
                    if len(val) == 1:
                        val.append(False)
                    else:
                        val[1] = to_bool(val[1])

                    watcher['hooks'][hook_name] = val
                # default bool to True
                elif opt in ('check_flapping', 'respawn', 'autostart',
                             'close_child_stdin'):
                    watcher[opt] = dget(section, opt, True, bool)
                else:
                    # freeform
                    watcher[opt] = val

            if watcher['copy_env']:
                watcher['env'] = dict(global_env)
            else:
                watcher['env'] = dict(local_env)

            watchers.append(watcher)

    # making sure we return consistent lists
    name = operator.itemgetter('name')
    watchers.sort(key=name)
    plugins.sort(key=name)
    sockets.sort(key=name)

    # Second pass to make sure env sections apply to all watchers.

    def _extend(target, source):
        for name, value in source:
            if name in target:
                continue
            target[name] = value

    def _expand_vars(target, key, env):
        if isinstance(target[key], str):
            target[key] = replace_gnu_args(target[key], env=env)
        elif isinstance(target[key], dict):
            for k in target[key].keys():
                _expand_vars(target[key], k, env)

    def _expand_section(section, env, exclude=None):
        if exclude is None:
            exclude = ('name', 'env')

        for option in section.keys():
            if option in exclude:
                continue
            _expand_vars(section, option, env)

    # build environment for watcher sections
    for section in cfg.sections():
        if section.startswith('env:'):
            section_elements = section.split("env:", 1)[1]
            watcher_patterns = [s.strip() for s in section_elements.split(',')]
            env_items = dict(cfg.items(section, noreplace=True))

            for pattern in watcher_patterns:
                match = [w for w in watchers if fnmatch(w['name'], pattern)]

                for watcher in match:
                    watcher['env'].update(env_items)

    # expand environment for watcher sections
    for watcher in watchers:
        env = dict(global_env)
        env.update(watcher['env'])
        _expand_section(watcher, env)

    config['watchers'] = watchers
    config['plugins'] = plugins
    config['sockets'] = sockets
    return config



from circus.commands.base import Command
from circus.commands.util import convert_option, validate_option
from circus.exc import ArgumentError, MessageError


class Set(Command):
    """\
        Set a watcher option
        ====================

        ZMQ Message
        -----------

        ::

            {
                "command": "set",
                "properties": {
                    "name": "nameofwatcher",
                    "options": {
                        "key1": "val1",
                        ..
                    }
                    "waiting": False
                }
            }


        The response return the status "ok". See the command Options for
        a list of key to set.

        Command line
        ------------

        ::

            $ circusctl set <name> <key1> <value1> <key2> <value2> --waiting


    """

    name = "set"
    properties = ['name', 'options']
    options = Command.waiting_options

    def message(self, *args, **opts):
        if len(args) < 3:
            raise ArgumentError("Invalid number of arguments")

        args = list(args)
        watcher_name = args.pop(0)
        if len(args) % 2 != 0:
            raise ArgumentError("List of key/values is invalid")

        options = {}
        while len(args) > 0:
            kv, args = args[:2], args[2:]
            kvl = kv[0].lower()
            options[kvl] = convert_option(kvl, kv[1])

        if opts.get('waiting', False):
            return self.make_message(name=watcher_name, waiting=True,
                                     options=options)
        else:
            return self.make_message(name=watcher_name, options=options)

    def execute(self, arbiter, props):
        watcher = self._get_watcher(arbiter, props.pop('name'))
        action = 0
        for key, val in props.get('options', {}).items():
            if key == 'hooks':
                new_action = 0
                for name, _val in val.items():
                    action = watcher.set_opt('hooks.%s' % name, _val)
                    if action == 1:
                        new_action = 1
            else:
                new_action = watcher.set_opt(key, val)

            if new_action == 1:
                action = 1
        # trigger needed action
        return watcher.do_action(action)

    def validate(self, props):
        super(Set, self).validate(props)

        options = props['options']
        if not isinstance(options, dict):
            raise MessageError("'options' property should be an object")

        for key, val in options.items():
            validate_option(key, val)



from circus.commands.base import Command


class ReloadConfig(Command):
    """\
        Reload the configuration file
        =============================

        This command reloads the configuration file, so changes in the
        configuration file will be reflected in the configuration of
        circus.


        ZMQ Message
        -----------

        ::

            {
                "command": "reloadconfig",
                "waiting": False
            }

        The response return the status "ok". If the property graceful is
        set to true the processes will be exited gracefully.


        Command line
        ------------

        ::

            $ circusctl reloadconfig [--waiting]

    """
    name = "reloadconfig"
    options = Command.waiting_options

    def message(self, *args, **opts):
        return self.make_message(**opts)

    def execute(self, arbiter, props):
        return arbiter.reload_from_config()





class AlreadyExist(Exception):
    """Raised when a watcher exists """
    pass


class MessageError(Exception):
    """ error raised when a message is invalid """
    pass


class CallError(Exception):
    pass


class ArgumentError(Exception):
    """Exception raised when one argument or the number of
    arguments is invalid"""
    pass


class ConflictError(Exception):
    """Exception raised when one exclusive command is already running
    in background"""
    pass



from circus.commands.base import Command
from circus.commands.restart import execute_watcher_start_stop_restart
from circus.commands.restart import match_options
from circus.exc import ArgumentError


class Start(Command):
    """\
        Start the arbiter or a watcher
        ==============================

        This command starts all the processes in a watcher or all watchers.


        ZMQ Message
        -----------

        ::

            {
                "command": "start",
                "properties": {
                    "name": '<name>",
                    "waiting": False,
                    "match": "[simple|glob|regex]"
                }
            }

        The response return the status "ok".

        If the property name is present, the watcher will be started.

        If ``waiting`` is False (default), the call will return immediately
        after calling `start` on each process.

        If ``waiting`` is True, the call will return only when the start
        process is completely ended. Because of the
        :ref:`graceful_timeout option <graceful_timeout>`, it can take some
        time.

        The ``match`` parameter can have the value ``simple`` for string
        compare, ``glob`` for wildcard matching (default) or ``regex`` for
        regex matching.


        Command line
        ------------

        ::

        $ circusctl restart [name] [--waiting] [--match=simple|glob|regex]

        Options
        +++++++

        - <name>: name or pattern of the watcher(s)
        - <match>: watcher match method

    """
    name = "start"
    options = list(Command.waiting_options)
    options.append(match_options)

    def message(self, *args, **opts):
        if len(args) > 1:
            raise ArgumentError("Invalid number of arguments")

        if len(args) == 1:
            return self.make_message(name=args[0], **opts)

        return self.make_message(**opts)

    def execute(self, arbiter, props):
        return execute_watcher_start_stop_restart(
            self, arbiter, props, 'start', arbiter.start_watchers,
            arbiter.start_watchers)



import signal
import traceback
import sys

from circus import logger
from circus.client import make_json
from circus.util import IS_WINDOWS


class SysHandler(object):

    _SIGNALS_NAMES = ("ILL ABRT BREAK INT TERM" if IS_WINDOWS else
                      "HUP QUIT INT TERM WINCH")

    SIGNALS = [getattr(signal, "SIG%s" % x) for x in _SIGNALS_NAMES.split()]

    SIG_NAMES = dict(
        (getattr(signal, name), name[3:].lower()) for name in dir(signal)
        if name[:3] == "SIG" and name[3] != "_"
    )

    def __init__(self, controller):
        self.controller = controller

        # init signals
        logger.info('Registering signals...')
        self._old = {}
        self._register()

    def stop(self):
        for sig, callback in self._old.items():
            try:
                signal.signal(sig, callback)
            except ValueError:
                pass

    def _register(self):
        for sig in self.SIGNALS:
            self._old[sig] = signal.getsignal(sig)
            signal.signal(sig, self.signal)

        # Don't let SIGQUIT and SIGUSR1 disturb active requests
        # by interrupting system calls
        if hasattr(signal, 'siginterrupt'):  # python >= 2.6
            signal.siginterrupt(signal.SIGQUIT, False)
            signal.siginterrupt(signal.SIGUSR1, False)

    def signal(self, sig, frame=None):
        signame = self.SIG_NAMES.get(sig)
        logger.info('Got signal SIG_%s' % signame.upper())

        if signame is not None:
            try:
                handler = getattr(self, "handle_%s" % signame)
                handler()
            except AttributeError:
                pass
            except Exception as e:
                tb = traceback.format_exc()
                logger.error("error: %s [%s]" % (e, tb))
                sys.exit(1)

    def quit(self):
        # We need to transfer the control to the loop's thread
        self.controller.loop.add_callback_from_signal(
            self.controller.dispatch, (None, make_json("quit"))
        )

    def reload(self):
        # We need to transfer the control to the loop's thread
        self.controller.loop.add_callback_from_signal(
            self.controller.dispatch,
            (None, make_json("reload", graceful=True))
        )

    def handle_int(self):
        self.quit()

    def handle_term(self):
        self.quit()

    def handle_quit(self):
        self.quit()

    def handle_ill(self):
        self.quit()

    def handle_abrt(self):
        self.quit()

    def handle_break(self):
        self.quit()

    def handle_winch(self):
        pass

    def handle_hup(self):
        self.reload()



import sys
import argparse
import os

from circus.config import get_config

try:
    import resource
except ImportError:
    resource = None     # NOQA

from circus import logger
from circus.arbiter import Arbiter
from circus.pidfile import Pidfile
from circus import __version__
from circus.util import MAXFD, REDIRECT_TO, configure_logger, LOG_LEVELS
from circus.util import check_future_exception_and_log


def get_maxfd():
    if not resource:
        maxfd = MAXFD
    else:
        maxfd = resource.getrlimit(resource.RLIMIT_NOFILE)[1]
        if maxfd == resource.RLIM_INFINITY:
            maxfd = MAXFD
    return maxfd


try:
    from os import closerange
except ImportError:
    def closerange(fd_low, fd_high):    # NOQA
        # Iterate through and close all file descriptors.
        for fd in range(fd_low, fd_high):
            try:
                os.close(fd)
            except OSError:    # ERROR, fd wasn't open to begin with (ignored)
                pass


# http://www.svbug.com/documentation/comp.unix.programmer-FAQ/faq_2.html#SEC16
def daemonize():
    """Standard daemonization of a process.
    """
    # guard to prevent daemonization with gevent loaded
    for module in sys.modules.keys():
        if module.startswith('gevent'):
            raise ValueError('Cannot daemonize if gevent is loaded')

    if hasattr(os, 'fork'):
        child_pid = os.fork()
    else:
        raise ValueError("Daemonizing is not available on this platform.")

    if child_pid != 0:
        # we're in the parent
        os._exit(0)

    # child process
    os.setsid()

    subchild = os.fork()
    if subchild:
        os._exit(0)

    # subchild
    maxfd = get_maxfd()
    closerange(0, maxfd)

    os.open(REDIRECT_TO, os.O_RDWR)
    os.dup2(0, 1)
    os.dup2(0, 2)


def main():
    import zmq
    try:
        zmq_version = [int(part) for part in zmq.__version__.split('.')[:2]]
        if len(zmq_version) < 2:
            raise ValueError()
    except (AttributeError, ValueError):
        print('Unknown PyZQM version - aborting...')
        sys.exit(0)

    if zmq_version[0] < 13 or (zmq_version[0] == 13 and zmq_version[1] < 1):
        print('circusd needs PyZMQ >= 13.1.0 to run - aborting...')
        sys.exit(0)

    parser = argparse.ArgumentParser(description='Run some watchers.')
    parser.add_argument('config', help='configuration file', nargs='?')

    # XXX we should be able to add all these options in the config file as well
    parser.add_argument('--log-level', dest='loglevel',
                        choices=list(LOG_LEVELS.keys()) + [
                            key.upper() for key in LOG_LEVELS.keys()],
                        help="log level")
    parser.add_argument('--log-output', dest='logoutput', help=(
        "The location where the logs will be written. The default behavior "
        "is to write to stdout (you can force it by passing '-' to "
        "this option). Takes a filename otherwise."))
    parser.add_argument("--logger-config", dest="loggerconfig", help=(
        "The location where a standard Python logger configuration INI, "
        "JSON or YAML file can be found.  This can be used to override "
        "the default logging configuration for the arbiter."))

    parser.add_argument('--daemon', dest='daemonize', action='store_true',
                        help="Start circusd in the background. Not supported "
                             "on Windows")
    parser.add_argument('--pidfile', dest='pidfile')
    parser.add_argument('--version', action='store_true', default=False,
                        help='Displays Circus version and exits.')

    args = parser.parse_args()

    if args.version:
        print(__version__)
        sys.exit(0)

    if args.config is None:
        parser.print_usage()
        sys.exit(0)

    if args.daemonize:
        daemonize()

    cfg = get_config(args.config)
    loglevel_from_conf = cfg.get('loglevel', None)
    logoutput_from_conf = cfg.get('logoutput', None)
    loggerconfig_from_conf = cfg.get('loggerconfig', None)

    # configure the logger
    loglevel = args.loglevel or loglevel_from_conf or 'info'
    logoutput = args.logoutput or logoutput_from_conf or '-'
    loggerconfig = args.loggerconfig or loggerconfig_from_conf or None
    configure_logger(logger, loglevel, logoutput, loggerconfig)

    # From here it can also come from the arbiter configuration
    # load the arbiter from config
    arbiter = Arbiter.load_from_config(args.config)

    # go ahead and set umask early if it is in the config
    if arbiter.umask is not None:
        os.umask(arbiter.umask)

    pidfile = args.pidfile or arbiter.pidfile or None
    if pidfile:
        pidfile = Pidfile(pidfile)

        try:
            pidfile.create(os.getpid())
        except RuntimeError as e:
            print(str(e))
            sys.exit(1)

    # Main loop
    restart = True
    while restart:
        try:
            arbiter = arbiter or Arbiter.load_from_config(args.config)
            future = arbiter.start()
            restart = False
            if check_future_exception_and_log(future) is None:
                restart = arbiter._restarting
        except Exception as e:
            # emergency stop
            arbiter.loop.run_sync(arbiter._emergency_stop)
            raise(e)
        except KeyboardInterrupt:
            pass
        finally:
            arbiter = None
            # Do not delete pid file if not going to exit
            if pidfile is not None and restart is False:
                pidfile.unlink()

    sys.exit(0)


if __name__ == '__main__':
    main()



from circus.commands.base import Command
from circus.exc import MessageError


class Listen(Command):
    """\
        Subscribe to a watcher event
        ============================

        ZMQ
        ---

        At any moment you can subscribe to a circus event. Circus provides
        a PUB/SUB feed on which any clients can subscribe. The subscriber
        endpoint URI is set in the circus.ini configuration file.

        Events are pubsub topics:

        - `watcher.<watchername>.reap`: when a process is reaped
        - `watcher.<watchername>.spawn`: when a process is spawned
        - `watcher.<watchername>.kill`: when a process is killed
        - `watcher.<watchername>.updated`: when watcher configuration
          is updated
        - `watcher.<watchername>.stop`: when a watcher is stopped
        - `watcher.<watchername>.start`: when a watcher is started

        All events messages are in a json struct.

        Command line
        ------------

        The client has been updated to provide a simple way to listen on the
        events::

            circusctl listen [<topic>, ...]

        Example of result:
        ++++++++++++++++++

        ::

            $ circusctl listen tcp://127.0.0.1:5556
            watcher.refuge.spawn: {u'process_id': 6, u'process_pid': 72976,
                                   u'time': 1331681080.985104}
            watcher.refuge.spawn: {u'process_id': 7, u'process_pid': 72995,
                                   u'time': 1331681086.208542}
            watcher.refuge.spawn: {u'process_id': 8, u'process_pid': 73014,
                                   u'time': 1331681091.427005}
    """
    name = "listen"
    msg_type = "sub"

    def message(self, *args, **opts):
        if not args:
            return [""]
        return list(args)

    def execute(self, arbiter, args):
        raise MessageError("invalid message. use a pub/sub socket")



from circus.commands.base import Command
from circus.exc import ArgumentError

from circus import logger


class List(Command):
    """\
        Get list of watchers or processes in a watcher
        ==============================================

        ZMQ Message
        -----------


        To get the list of all the watchers::

            {
                "command": "list",
            }


        To get the list of active processes in a watcher::

            {
                "command": "list",
                "properties": {
                    "name": "nameofwatcher",
                }
            }


        The response return the list asked. the mapping returned can either be
        'watchers' or 'pids' depending the request.

        Command line
        ------------

        ::

            $ circusctl list [<name>]
    """
    name = "list"

    def message(self, *args, **opts):
        if len(args) > 1:
            raise ArgumentError("Invalid number of arguments")

        if len(args) == 1:
            return self.make_message(name=args[0])
        else:
            return self.make_message()

    def execute(self, arbiter, props):
        if 'name' in props:
            watcher = self._get_watcher(arbiter, props['name'])
            processes = watcher.get_active_processes()
            status = [(p.pid, p.status) for p in processes]
            logger.debug('here is the status of the processes %s' % status)
            return {"pids":  [p.pid for p in processes]}
        else:
            watchers = sorted(arbiter._watchers_names)
            return {"watchers": [name for name in watchers]}

    def console_msg(self, msg):
        if "pids" in msg:
            return ",".join([str(process_id)
                             for process_id in msg.get('pids')])
        elif 'watchers' in msg:
            return ",".join([watcher for watcher in msg.get('watchers')])
        if 'reason' not in msg:
            msg['reason'] = "Response doesn't contain 'pids' nor 'watchers'."
        return self.console_error(msg)



from circus.commands.base import Command
from circus.exc import ArgumentError


class RmWatcher(Command):
    """\
        Remove a watcher
        ================

        This command removes a watcher dynamically from the arbiter. The
        watchers are gracefully stopped by default.

        ZMQ Message
        -----------

        ::

            {
                "command": "rm",
                "properties": {
                    "name": "<nameofwatcher>",
                    "nostop": False,
                    "waiting": False
                }
            }

        The response return a status "ok".

        If ``nostop`` is True (default: False), the processes for the watcher
        will not be stopped - instead the watcher will just be forgotten by
        circus and the watcher processes will be responsible for stopping
        themselves. If ``nostop`` is not specified or is False, then the
        watcher processes will be stopped gracefully.

        If ``waiting`` is False (default), the call will return immediately
        after starting to remove and stop the corresponding watcher.

        If ``waiting`` is True, the call will return only when the remove and
        stop process is completely ended. Because of the
        :ref:`graceful_timeout option <graceful_timeout>`, it can take some
        time.


        Command line
        ------------

        ::

            $ circusctl rm <name> [--waiting] [--nostop]

        Options
        +++++++

        - <name>: name of the watcher to remove
        - nostop: do not stop the watcher processes, just remove the watcher

    """

    name = "rm"
    properties = ['name']
    options = Command.waiting_options + \
        [('nostop', 'nostop', False, 'Do not stop watcher processes')]

    def message(self, *args, **opts):
        if len(args) < 1 or len(args) > 1:
            raise ArgumentError("Invalid number of arguments")

        return self.make_message(name=args[0])

    def execute(self, arbiter, props):
        self._get_watcher(arbiter, props['name'])
        return arbiter.rm_watcher(props['name'], props.get('nostop', False))



from circus.arbiter import Arbiter as _Arbiter
from circus.green.controller import Controller

from zmq.green.eventloop import ioloop
from zmq.green import Context


class Arbiter(_Arbiter):
    def _init_context(self, context):
        self.context = context or Context.instance()
        self.loop = ioloop.IOLoop.current()
        self.ctrl = Controller(self.endpoint, self.multicast_endpoint,
                               self.context, self.loop, self, self.check_delay)



from collections import defaultdict
from itertools import chain
import os
import errno
import socket

import zmq
import zmq.utils.jsonapi as json
from tornado import ioloop
from zmq.eventloop import zmqstream

from circus.commands import get_commands
from circus.client import CircusClient
from circus.stats.collector import WatcherStatsCollector, SocketStatsCollector
from circus.stats.publisher import StatsPublisher
from circus import logger
from circus.util import to_str


class StatsStreamer(object):
    def __init__(self, endpoint, pubsub_endoint, stats_endpoint,
                 ssh_server=None, delay=1., loop=None):
        self.topic = b'watcher.'
        self.delay = delay
        self.ctx = zmq.Context()
        self.pubsub_endpoint = pubsub_endoint
        self.sub_socket = self.ctx.socket(zmq.SUB)
        self.sub_socket.setsockopt(zmq.SUBSCRIBE, self.topic)
        self.sub_socket.connect(self.pubsub_endpoint)
        self.loop = loop or ioloop.IOLoop.current()
        self.substream = zmqstream.ZMQStream(self.sub_socket, self.loop)
        self.substream.on_recv(self.handle_recv)
        self.client = CircusClient(context=self.ctx, endpoint=endpoint,
                                   ssh_server=ssh_server)
        self.cmds = get_commands()
        self.publisher = StatsPublisher(stats_endpoint, self.ctx)
        self._initialize()

    def _initialize(self):
        self._pids = defaultdict(list)
        self._callbacks = dict()
        self.running = False  # should the streamer be running?
        self.stopped = False  # did the collect started yet?
        self.circus_pids = {}
        self.sockets = []
        self.get_watchers = self._pids.keys

    def get_pids(self, watcher=None):
        if watcher is not None:
            if watcher == 'circus':
                return list(self.circus_pids.keys())
            return self._pids[watcher]
        return chain(*list(self._pids.values()))

    def get_circus_pids(self):
        watchers = self.client.send_message('list').get('watchers', [])

        # getting the circusd, circusd-stats and circushttpd pids
        res = self.client.send_message('dstats')
        pids = {os.getpid(): 'circusd-stats'}

        if 'info' in res:
            pids[res['info']['pid']] = 'circusd'

        if 'circushttpd' in watchers:
            httpd_pids = self.client.send_message('list', name='circushttpd')

            if 'pids' in httpd_pids:
                httpd_pids = httpd_pids['pids']
                if len(httpd_pids) == 1:
                    pids[httpd_pids[0]] = 'circushttpd'

        return pids

    def _add_callback(self, name, start=True, kind='watcher'):
        logger.debug('Callback added for %s' % name)

        if kind == 'watcher':
            klass = WatcherStatsCollector
        elif kind == 'socket':
            klass = SocketStatsCollector
        else:
            raise ValueError('Unknown callback kind %r' % kind)

        self._callbacks[name] = klass(self, name, self.delay, self.loop)
        if start:
            self._callbacks[name].start()

    def _init(self):
        self._pids.clear()

        # getting the initial list of watchers/pids
        res = self.client.send_message('list')

        for watcher in res['watchers']:
            if watcher in ('circusd', 'circushttpd', 'circusd-stats'):
                # this is dealt by the special 'circus' collector
                continue

            pid_list = self.client.send_message('list', name=watcher)
            pids = pid_list.get('pids', [])
            for pid in pids:
                self._append_pid(watcher, pid)

        # getting the circus pids
        self.circus_pids = self.get_circus_pids()
        if 'circus' not in self._callbacks:
            self._add_callback('circus')
        else:
            self._callbacks['circus'].start()

        # getting the initial list of sockets
        res = self.client.send_message('listsockets')
        for sock in res.get('sockets', []):
            fd = sock['fd']
            if 'path' in sock:
                # unix socket
                address = sock['path']
            else:
                address = '%s:%s' % (sock['host'], sock['port'])

            # XXX type / family ?
            sock = socket.fromfd(fd, socket.AF_INET, socket.SOCK_STREAM)
            self.sockets.append((sock, address, fd))

        self._add_callback('sockets', kind='socket')

    def stop_watcher(self, watcher):
        for pid in self._pids[watcher]:
            self.remove_pid(watcher, pid)

    def remove_pid(self, watcher, pid):
        if pid in self._pids[watcher]:
            logger.debug('Removing %d from %s' % (pid, watcher))
            self._pids[watcher].remove(pid)
            if len(self._pids[watcher]) == 0:
                logger.debug(
                    'Stopping the periodic callback for {0}' .format(watcher))
                self._callbacks[watcher].stop()

    def _append_pid(self, watcher, pid):
        if watcher not in self._pids or len(self._pids[watcher]) == 0:
            logger.debug(
                'Starting the periodic callback for {0}'.format(watcher))
            if watcher not in self._callbacks:
                self._add_callback(watcher)
            else:
                self._callbacks[watcher].start()

        if pid in self._pids[watcher]:
            return
        self._pids[watcher].append(pid)
        logger.debug('Adding %d in %s' % (pid, watcher))

    def start(self):
        self.running = True
        logger.info('Starting the stats streamer')
        self._init()
        logger.debug('Initial list is ' + str(self._pids))
        logger.debug('Now looping to get circusd events')

        while self.running:
            try:
                self.loop.start()
            except zmq.ZMQError as e:
                logger.debug(str(e))

                if e.errno == errno.EINTR:
                    continue
                elif e.errno == zmq.ETERM:
                    break
                else:
                    logger.debug("got an unexpected error %s (%s)", str(e),
                                 e.errno)
                    raise
            else:
                break
        self.stop()

    def handle_recv(self, data):
        """called each time circusd sends an event"""
        # maintains a periodic callback to compute mem and cpu consumption for
        # each pid.
        logger.debug('Received an event from circusd: %s' % str(data))
        topic, msg = data
        try:
            topic = to_str(topic)
            watcher = topic.split('.')[1:-1][0]
            action = topic.split('.')[-1]
            msg = json.loads(msg)

            if action in ('reap', 'kill'):
                # a process was reaped
                pid = msg['process_pid']
                self.remove_pid(watcher, pid)
            elif action == 'spawn':
                # a process was added
                pid = msg['process_pid']
                self._append_pid(watcher, pid)
            elif action == 'stop':
                # the whole watcher was stopped.
                self.stop_watcher(watcher)
            else:
                logger.debug('Unknown action: %r' % action)
                logger.debug(msg)
        except Exception:
            logger.exception('Failed to handle %r' % msg)

    def stop(self):
        # stop all the periodic callbacks running
        for callback in self._callbacks.values():
            callback.stop()

        self.loop.stop()
        self.ctx.destroy(0)
        self.publisher.stop()
        self.stopped = True
        self.running = False
        logger.info('Stats streamer stopped')



import re
import socket
import time

from tornado.ioloop import IOLoop

from circus.plugins import CircusPlugin
from circus import logger
from circus.util import to_signum, AsyncPeriodicCallback


class WatchDog(CircusPlugin):
    """Plugin that bind an udp socket and wait for watchdog messages.
    For "watchdoged" processes, the watchdog will kill them (using the "kill"
    command) if they don't send heartbeat in a certain period of time
    materialized by loop_rate * max_count. (circus will automatically restart
    the missing processes in the watcher)

    Each monitored process should send udp message at least at the loop_rate.
    The udp message format is a line of text, decoded using **msg_regex**
    parameter.
    The heartbeat message MUST at least contain the pid of the process sending
    the message.

    The list of monitored watchers are determined by the parameter
    **watchers_regex** in the configuration.

    At startup, the plugin does not know all the circus watchers and pids,
    so it's needed to discover all watchers and pids. After the discover, the
    monitoring list is updated by messages from circusd handled in
    self.handle_recv

    Plugin Options --

    - **loop_rate** -- watchdog loop rate in seconds. At each loop, WatchDog
      will looks for "dead" processes.
    - **watchers_regex** -- regex for matching watcher names that should be
      monitored by the watchdog (default: ".*" all watchers are monitored)
    - **msg_regex** -- regex for decoding the received heartbeat
      message in udp (default: "^(?P<pid>.*);(?P<timestamp>.*)$")
      the default format is a simple text message: "pid;timestamp"
    - **max_count** -- max number of passed loop without receiving
      any heartbeat before restarting process (default: 3)
    - **ip** -- ip the watchdog will bind on (default: 127.0.0.1)
    - **port** -- port the watchdog will bind on (default: 1664)
    - **watchers_stop_signal** -- optionally override the stop_signal used
      when killing the processes
    - **watchers_graceful_timeout** -- optionally override the graceful_timeout
      used when killing the processes
    """
    name = 'watchdog'

    def __init__(self, endpoint, pubsub_endpoint, check_delay, ssh_server,
                 **config):
        super(WatchDog, self).__init__(endpoint, pubsub_endpoint,
                                       check_delay, ssh_server=ssh_server)

        self.loop_rate = float(config.get("loop_rate", 60))  # in seconds
        self.watchers_regex = config.get("watchers_regex", ".*")
        self.msg_regex = config.get("msg_regex",
                                    "^(?P<pid>.*);(?P<timestamp>.*)$")
        self.max_count = int(config.get("max_count", 3))
        self.watchdog_ip = config.get("ip", "127.0.0.1")
        self.watchdog_port = int(config.get("port", 1664))
        self.stop_signal = config.get("watchers_stop_signal")
        if self.stop_signal:
            self.stop_signal = to_signum(self.stop_signal)
        self.graceful_timeout = config.get("watchers_graceful_timeout")
        if self.graceful_timeout:
            self.graceful_timeout = float(self.graceful_timeout)

        self.pid_status = dict()
        self.period = None
        self.starting = True

    def handle_init(self):
        """Initialization of plugin

        - set the periodic call back for the process monitoring (at loop_rate)
        - create the listening UDP socket
        """
        self.period = AsyncPeriodicCallback(self.look_after,
                                            self.loop_rate * 1000)
        self.period.start()
        self._bind_socket()

    def handle_stop(self):
        if self.period is not None:
            self.period.stop()
        self.sock.close()
        self.sock = None

    def handle_recv(self, data):
        """Handle received message from circusd

        We need to handle two messages:
        - spawn: add a new monitored child pid
        - reap: remove a killed child pid from monitoring
        """
        watcher_name, action, msg = self.split_data(data)
        logger.debug("received data from circusd: watcher.%s.%s, %s",
                     watcher_name, action, msg)
        # check if monitored watchers:
        if self._match_watcher_name(watcher_name):
            try:
                message = self.load_message(msg)
            except ValueError:
                logger.error("Error while decoding json for message: %s",
                             msg)
            else:
                if "process_pid" not in message:
                    logger.warning('no process_pid in message')
                    return
                pid = str(message.get("process_pid"))
                if action == "spawn":
                    self.pid_status[pid] = dict(watcher=watcher_name,
                                                last_activity=time.time())
                    logger.info("added new monitored pid for %s:%s",
                                watcher_name,
                                pid)
                # very questionable fix for Py3 here!
                # had to add check for pid in self.pid_status
                elif action == "reap" and pid in self.pid_status:
                    old_pid = self.pid_status.pop(pid)
                    logger.info("removed monitored pid for %s:%s",
                                old_pid['watcher'],
                                pid)

    def _discover_monitored_pids(self):
        """Try to discover all the monitored pids.

        This should be done only at startup time, because if new watchers or
        pids are created in running time, we should receive the message
        from circusd which is handled by self.handle_recv
        """
        self.pid_status = dict()
        all_watchers = self.call("list")
        for watcher_name in all_watchers['watchers']:
            if self._match_watcher_name(watcher_name):
                processes = self.call("list", name=watcher_name)
                if 'pids' in processes:
                    for pid in processes['pids']:
                        pid = str(pid)
                        self.pid_status[pid] = dict(watcher=watcher_name,
                                                    last_activity=time.time())
                        logger.info("discovered: %s, pid:%s",
                                    watcher_name,
                                    pid)

    def _bind_socket(self):
        """bind the listening socket for watchdog udp and start an event
        handler for handling udp received messages.
        """
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            self.sock.bind((self.watchdog_ip, self.watchdog_port))
        except socket.error as socket_error:
            logger.error(
                "Problem while binding watchdog socket on %s:%s (err %s",
                self.watchdog_ip,
                self.watchdog_port,
                str(socket_error))
            self.sock = None
        else:
            self.sock.settimeout(1)
            self.loop.add_handler(self.sock.fileno(),
                                  self.receive_udp_socket,
                                  IOLoop.READ)
            logger.info("Watchdog listening UDP on %s:%s",
                        self.watchdog_ip, self.watchdog_port)

    def _match_watcher_name(self, name):
        """Match the given watcher name with the watcher_regex given in config

        :return: re.match object or None
        """
        return re.match(self.watchers_regex, name)

    def _decode_received_udp_message(self, data):
        """decode the received message according to the msg_regex

        :return: decoded message
        :rtype: dict or None
        """
        result = re.match(self.msg_regex, data.decode())
        if result is not None:
            return result.groupdict()

    def receive_udp_socket(self, fd, events):
        """Check the socket for received UDP message.
        This method is periodically called by the ioloop.
        If messages are received and parsed, update the status of
        the corresponing pid.
        """
        data, _ = self.sock.recvfrom(1024)
        heartbeat = self._decode_received_udp_message(data)
        if "pid" in heartbeat:
            if heartbeat['pid'] in self.pid_status:
                # TODO: check and compare received time
                # with our own time.time()
                self.pid_status[heartbeat["pid"]][
                    'last_activity'] = time.time()
            else:
                logger.warning("received watchdog for a "
                               "non monitored process:%s",
                               heartbeat)
        logger.debug("watchdog message: %s", heartbeat)

    def look_after(self):
        """Checks for the watchdoged watchers and restart a process if no
        received watchdog after the loop_rate * max_count period.
        """
        # if first check, do a full discovery first.
        if self.starting:
            self._discover_monitored_pids()
            self.starting = False

        max_timeout = self.loop_rate * self.max_count
        too_old_time = time.time() - max_timeout
        for pid, detail in self.pid_status.items():
            if detail['last_activity'] < too_old_time:
                logger.info("watcher:%s, pid:%s is not responding. Kill it !",
                            detail['watcher'],
                            pid)

                props = dict(name=detail['watcher'], pid=int(pid))
                if self.stop_signal is not None:
                    props['signum'] = self.stop_signal
                if self.graceful_timeout is not None:
                    props['graceful_timeout'] = self.graceful_timeout

                self.cast('kill', **props)

                # Trusting watcher to eventually stop the process after
                # graceful timeout
                del self.pid_status[pid]



from circus.commands.base import Command
from circus.exc import ArgumentError, MessageError
from circus.util import to_signum


class Signal(Command):
    """\
        Send a signal
        =============

        This command allows you to send a signal to all processes in a watcher,
        a specific process in a watcher or its children.

        ZMQ Message
        -----------

        To send a signal to all the processes for a watcher::

            {
                "command": "signal",
                "property": {
                    "name": <name>,
                    "signum": <signum>
            }

        To send a signal to a process::

            {
                "command": "signal",
                "property": {
                    "name": <name>,
                    "pid": <processid>,
                    "signum": <signum>
            }

        An optional property "children" can be used to send the signal
        to all the children rather than the process itself::

            {
                "command": "signal",
                "property": {
                    "name": <name>,
                    "pid": <processid>,
                    "signum": <signum>,
                    "children": True
            }

        To send a signal to a process child::

            {
                "command": "signal",
                "property": {
                    "name": <name>,
                    "pid": <processid>,
                    "signum": <signum>,
                    "child_pid": <childpid>,
            }

        It is also possible to send a signal to all the children of the
        watcher::

            {
                "command": "signal",
                "property": {
                    "name": <name>,
                    "signum": <signum>,
                    "children": True
            }

        Lastly, you can send a signal to the process *and* its children, with
        the *recursive* option::

            {
                "command": "signal",
                "property": {
                    "name": <name>,
                    "signum": <signum>,
                    "recursive": True
            }



        Command line
        ------------

        ::

            $ circusctl signal <name> [<pid>] [--children]
                    [--recursive] <signum>

        Options:
        ++++++++

        - <name>: the name of the watcher
        - <pid>: integer, the process id.
        - <signum>: the signal number (or name) to send.
        - <childpid>: the pid of a child, if any
        - <children>: boolean, send the signal to all the children
        - <recursive>: boolean, send the signal to the process and its children

    """

    name = "signal"
    options = [('', 'children', False, "Only signal children of the process"),
               ('', 'recursive', False, "Signal parent and children")]
    properties = ['name', 'signum']

    def message(self, *args, **opts):
        largs = len(args)
        if largs < 2 or largs > 3:
            raise ArgumentError("Invalid number of arguments")

        props = {
            'name': args[0],
            'children': opts.get("children", False),
            'recursive': opts.get("recursive", False),
            'signum': args[-1],
        }
        if len(args) == 3:
            props['pid'] = int(args[1])
        return self.make_message(**props)

    def execute(self, arbiter, props):
        name = props.get('name')
        watcher = self._get_watcher(arbiter, name)
        signum = props.get('signum')
        pids = [props['pid']] if 'pid' in props else watcher.get_active_pids()
        childpid = props.get('childpid', None)
        children = props.get('children', False)
        recursive = props.get('recursive', False)

        for pid in pids:
            if childpid:
                watcher.send_signal_child(pid, childpid, signum)
            elif children:
                watcher.send_signal_children(pid, signum)
            else:
                # send to the given pid
                watcher.send_signal(pid, signum)

                if recursive:
                    # also send to the children
                    watcher.send_signal_children(pid, signum, recursive=True)

    def validate(self, props):
        super(Signal, self).validate(props)

        if 'childpid' in props and 'pid' not in props:
            raise ArgumentError('cannot specify childpid without pid')

        try:
            props['signum'] = to_signum(props['signum'])
        except ValueError:
            raise MessageError('signal invalid')



import threading
from threading import _active_limbo_lock, _active, _sys


debugger = False
try:
    # noinspection PyUnresolvedReferences
    import pydevd
    debugger = pydevd.GetGlobalDebugger()
except ImportError:
    pass

if not debugger:
    # see http://bugs.python.org/issue1596321
    if hasattr(threading.Thread, '_Thread__delete'):
        def _delete(self):
            try:
                with _active_limbo_lock:
                    del _active[self._Thread__ident]
            except KeyError:
                if 'dummy_threading' not in _sys.modules:
                    raise

        threading.Thread._Thread__delete = _delete
    else:
        def _delete(self):  # NOQA
            try:
                with _active_limbo_lock:
                    del _active[self._ident]
            except KeyError:
                if 'dummy_threading' not in _sys.modules:
                    raise

        threading.Thread._delete = _delete



from circus.commands.base import Command
from circus.exc import ArgumentError


class NumWatchers(Command):
    """\
        Get the number of watchers
        ==========================

        Get the number of watchers in a arbiter

        ZMQ Message
        -----------

        ::

            {
                "command": "numwatchers",
            }

        The response return the number of watchers in the 'numwatchers`
        property::

            { "status": "ok", "numwatchers": <n>, "time", "timestamp" }


        Command line
        ------------

        ::

            $ circusctl numwatchers

    """
    name = "numwatchers"

    def message(self, *args, **opts):
        if len(args) > 0:
            raise ArgumentError("Invalid number of arguments")
        return self.make_message()

    def execute(self, arbiter, props):
        return {"numwatchers": arbiter.numwatchers()}

    def console_msg(self, msg):
        if msg.get("status") == "ok":
            return str(msg.get("numwatchers"))
        return self.console_error(msg)



from circus.util import AsyncPeriodicCallback
from circus.controller import Controller as _Controller
from circus.green.sighandler import SysHandler

from zmq.green.eventloop import zmqstream


class Controller(_Controller):

    def _init_syshandler(self):
        self.sys_hdl = SysHandler(self)

    def _init_stream(self):
        self.stream = zmqstream.ZMQStream(self.ctrl_socket, self.loop)
        self.stream.on_recv(self.handle_message)

    def start(self):
        self.loop.make_current()
        self.initialize()
        self.caller = AsyncPeriodicCallback(self.arbiter.manage_watchers,
                                            self.check_delay)
        self.caller.start()



from circus.commands.base import Command
from circus.exc import ArgumentError, MessageError
from circus.util import to_signum
from tornado import gen


class Kill(Command):
    """\
        Kill a specific process
        =======================

        This command allows you to terminate a process in a watcher. If a
        process does not exit within graceful_timeout it will be terminated
        with a SIGKILL.

        ZMQ Message
        -----------

        To kill all the processes of a watcher::

            {
                "command": "kill",
                "property": {
                    "name": <name>,
                    "signum": <signum>,
                    "graceful_timeout": <graceful_timeout>
                }
            }

        To send a signal to a specific process::

            {
                "command": "kill",
                "property": {
                    "name": <name>,
                    "pid": <processid>,
                    "signum": <signum>,
                    "graceful_timeout": <graceful_timeout>
                }
            }


        Command line
        ------------

        ::

            $ circusctl kill <name> [<pid>] [--signum] [--graceful_timeout]

        Options:
        ++++++++

        - <name>: the name of the watcher
        - <pid>: integer, the process id
        - <signum>: overrides the watcher's stop_signal (number or name)
        - <graceful_timeout>: overrides the watcher's graceful_timeout

    """

    name = "kill"
    properties = ['name']
    options = Command.waiting_options + [
        ('signum', 'signum', None, "overrides the watcher's stop_signal "
                                   "(number or name)"),
        ('graceful_timeout', 'graceful_timeout', None, "overrides the watcher"
                                                       "'s graceful_timeout"),
    ]

    def message(self, *args, **opts):
        if len(args) < 1 or len(args) > 2:
            raise ArgumentError("Invalid number of arguments")

        props = {}
        props['name'] = args[0]
        if len(args) > 1:
            props['pid'] = args[1]
        if opts['signum'] is not None:
            props['signum'] = opts['signum']
        if opts['graceful_timeout'] is not None:
            props['graceful_timeout'] = opts['graceful_timeout']
        return self.make_message(**props)

    @gen.coroutine
    def execute(self, arbiter, props):
        name = props.get('name')
        pid = props.get('pid')
        signum = props.get('signum')
        graceful_timeout = props.get('graceful_timeout')

        watcher = self._get_watcher(arbiter, name)
        processes = watcher.get_active_processes()
        if pid:
            processes = [p for p in processes if p.pid == pid]

        if processes:
            yield [watcher.kill_process(p,
                                        stop_signal=signum,
                                        graceful_timeout=graceful_timeout)
                   for p in processes]

    def validate(self, props):
        super(Kill, self).validate(props)

        if 'pid' in props:
            props['pid'] = int(props['pid'])

        try:
            if 'signum' in props:
                props['signum'] = to_signum(props['signum'])
        except ValueError:
            raise MessageError('signal invalid')



from circus.commands.incrproc import IncrProc
from circus.util import TransformableFuture


class DecrProc(IncrProc):
    """\
        Decrement the number of processes in a watcher
        ==============================================

        This comment decrement the number of processes in a watcher
        by <nbprocess>, 1 being the default.

        ZMQ Message
        -----------

        ::

            {
                "command": "decr",
                "propeties": {
                    "name": "<watchername>"
                    "nb": <nbprocess>
                    "waiting": False
                }
            }

        The response return the number of processes in the 'numprocesses`
        property::

            { "status": "ok", "numprocesses": <n>, "time", "timestamp" }

        Command line
        ------------

        ::

            $ circusctl decr <name> [<nb>] [--waiting]

        Options
        +++++++

        - <name>: name of the watcher
        - <nb>: the number of processes to remove.

    """
    name = "decr"
    properties = ['name']

    def execute(self, arbiter, props):
        watcher = self._get_watcher(arbiter, props.get('name'))
        if watcher.singleton:
            return {"numprocesses": watcher.numprocesses, "singleton": True}
        else:
            nb = props.get('nb', 1)
            resp = TransformableFuture()
            resp.set_upstream_future(watcher.decr(nb))
            resp.set_transform_function(lambda x: {"numprocesses": x})
            return resp



import copy
import textwrap
import time

from circus.exc import MessageError
from circus.commands import errors


KNOWN_COMMANDS = []


def get_commands():
    commands = {}
    for c in KNOWN_COMMANDS:
        cmd = c()
        commands[c.name] = cmd.copy()
    return commands


def ok(props=None):
    resp = {"status": "ok", "time": time.time()}
    if props:
        resp.update(props)
    return resp


def error(reason="unknown", tb=None, errno=errors.NOT_SPECIFIED):
    return {
        "status": "error",
        "reason": reason,
        "tb": tb,
        "time": time.time(),
        "errno": errno
    }


class CommandMeta(type):

    def __new__(cls, name, bases, attrs):
        super_new = type.__new__
        parents = [b for b in bases if isinstance(b, CommandMeta)]

        if not parents:
            return super_new(cls, name, bases, attrs)

        attrs["order"] = len(KNOWN_COMMANDS)
        new_class = super_new(cls, name, bases, attrs)
        new_class.fmt_desc()
        KNOWN_COMMANDS.append(new_class)
        return new_class

    def fmt_desc(cls):
        desc = textwrap.dedent(cls.__doc__).strip()
        setattr(cls, "desc",  desc)
        setattr(cls, "short", desc.splitlines()[0])


class Command(object):

    name = None
    msg_type = "dealer"
    options = []
    properties = []
    waiting = False
    waiting_options = [('waiting', 'waiting', False,
                        "Waiting the real end of the process")]

    ##################################################
    # These methods run within the circusctl process #
    ##################################################

    def make_message(self, **props):
        name = props.pop("command", self.name)
        return {"command": name, "properties": props or {}}

    def message(self, *args, **opts):
        raise NotImplementedError("message function isn't implemented")

    def console_error(self, msg):
        return "error: %s" % msg.get("reason")

    def console_msg(self, msg):
        if msg.get('status') == "ok":
            return "ok"
        return self.console_error(msg)

    def copy(self):
        return copy.copy(self)

    ################################################
    # These methods run within the circusd process #
    ################################################

    def execute(self, arbiter, props):
        raise NotImplementedError("execute function is not implemented")

    def _get_watcher(self, arbiter, watcher_name):
        """Get watcher from the arbiter if any."""
        try:
            return arbiter.get_watcher(watcher_name.lower())
        except KeyError:
            raise MessageError("program %s not found" % watcher_name)

    def validate(self, props):
        if not self.properties:
            return

        for propname in self.properties:
            if propname not in props:
                raise MessageError("message invalid %r is missing" % propname)


Command = CommandMeta('Command', (Command,), {})



import re
import fnmatch
from functools import partial
from circus.commands.base import Command
from circus.exc import ArgumentError, MessageError
from circus.util import TransformableFuture


def execute_watcher_start_stop_restart(command, arbiter, props,
                                       watcher_function_name,
                                       watchers_function, arbiter_function):
    """base function to handle start/stop/restart watcher requests.
    since this is always the same procedure except some function names this
    function handles all watcher start/stop commands
    """
    if 'name' in props:
        match = props.get('match', 'glob')
        if match == 'simple':
            watchers = [command._get_watcher(arbiter, props['name'])]
        else:
            watcher_name = props['name'].lower()
            if match == 'glob':
                name = re.compile(fnmatch.translate(watcher_name))
            elif match == 'regex':
                name = re.compile(watcher_name)
            else:
                raise MessageError("unknown match method %s" % match)
            watchers = [watcher
                        for watcher in arbiter.iter_watchers()
                        if name.match(watcher.name.lower())]

        if not watchers:
            raise MessageError("program %s not found" % props['name'])

        if len(watchers) == 1:
            if props.get('waiting'):
                resp = TransformableFuture()
                func = getattr(watchers[0], watcher_function_name)
                resp.set_upstream_future(func())
                resp.set_transform_function(lambda x: {"info": x})
                return resp
            return getattr(watchers[0], watcher_function_name)()

        def watcher_iter_func(reverse=True):
            return sorted(watchers, key=lambda a: a.priority, reverse=reverse)

        return watchers_function(watcher_iter_func=watcher_iter_func)
    else:
        return arbiter_function()


match_options = ('match', 'match', 'glob',
                 "Watcher name matching method (simple, glob or regex)")


class Restart(Command):
    """\
        Restart the arbiter or a watcher
        ================================

        This command restart all the process in a watcher or all watchers. This
        funtion simply stop a watcher then restart it.

        ZMQ Message
        -----------

        ::

            {
                "command": "restart",
                "properties": {
                    "name": "<name>",
                    "waiting": False,
                    "match": "[simple|glob|regex]"
                }
            }

        The response return the status "ok".

        If the property name is present, then the reload will be applied
        to the watcher.

        If ``waiting`` is False (default), the call will return immediately
        after calling `stop_signal` on each process.

        If ``waiting`` is True, the call will return only when the restart
        process is completely ended. Because of the
        :ref:`graceful_timeout option <graceful_timeout>`, it can take some
        time.

        The ``match`` parameter can have the value ``simple`` for string
        compare, ``glob`` for wildcard matching (default) or ``regex`` for
        regex matching.


        Command line
        ------------

        ::

            $ circusctl restart [name] [--waiting] [--match=simple|glob|regex]

        Options
        +++++++

        - <name>: name or pattern of the watcher(s)
        - <match>: watcher match method
    """

    name = "restart"
    options = list(Command.waiting_options)
    options.append(match_options)

    def message(self, *args, **opts):
        if len(args) > 1:
            raise ArgumentError("Invalid number of arguments")

        if len(args) == 1:
            return self.make_message(name=args[0], **opts)

        return self.make_message(**opts)

    def execute(self, arbiter, props):
        return execute_watcher_start_stop_restart(
            self, arbiter, props, 'restart', arbiter.restart,
            partial(arbiter.restart, inside_circusd=True))



import os
import sys
import traceback
import functools
from queue import Queue, Empty  # noqa: F401
from urllib.parse import urlparse


import zmq
import zmq.utils.jsonapi as json
from tornado import ioloop
from zmq.eventloop import zmqstream
from tornado.concurrent import Future

from circus.util import create_udp_socket
from circus.util import check_future_exception_and_log
from circus.util import to_uid
from circus.util import AsyncPeriodicCallback
from circus.commands import get_commands, ok, error, errors
from circus import logger
from circus.exc import MessageError, ConflictError
from circus.sighandler import SysHandler


class Controller(object):

    def __init__(self, endpoint, multicast_endpoint, context, loop, arbiter,
                 check_delay=1.0, endpoint_owner=None):
        self.arbiter = arbiter
        self.caller = None
        self.endpoint = endpoint
        self.multicast_endpoint = multicast_endpoint
        self.context = context
        self.loop = loop
        self.check_delay = check_delay * 1000
        self.endpoint_owner = endpoint_owner
        self.started = False

        # initialize the sys handler
        self._init_syshandler()

        # get registered commands
        self.commands = get_commands()

    def _init_syshandler(self):
        self.sys_hdl = SysHandler(self)

    def _init_stream(self):
        self.stream = zmqstream.ZMQStream(self.ctrl_socket, self.loop)
        self.stream.on_recv(self.handle_message)

    def _init_multicast_endpoint(self):
        multicast_addr, multicast_port = urlparse(self.multicast_endpoint)\
            .netloc.split(':')
        try:
            self.udp_socket = create_udp_socket(multicast_addr,
                                                multicast_port)
            self.loop.add_handler(self.udp_socket.fileno(),
                                  self.handle_autodiscover_message,
                                  ioloop.IOLoop.READ)
        except (IOError, OSError, ValueError):
            message = ("Multicast discovery is disabled, there was an "
                       "error during udp socket creation.")
            logger.warning(message, exc_info=True)

    @property
    def endpoint_owner_mode(self):
        return self.endpoint_owner is not None and \
            self.endpoint.startswith('ipc://')

    def initialize(self):
        # initialize controller

        # Initialize ZMQ Sockets
        self.ctrl_socket = self.context.socket(zmq.ROUTER)
        self.ctrl_socket.bind(self.endpoint)
        self.ctrl_socket.linger = 0

        # support chown'ing the zmq endpoint on unix platforms
        if self.endpoint_owner_mode:
            uid = to_uid(self.endpoint_owner)
            sockpath = self.endpoint[6:]  # length of 'ipc://' prefix
            os.chown(sockpath, uid, -1)

        self._init_stream()

        # Initialize UDP Socket
        if self.multicast_endpoint:
            self._init_multicast_endpoint()

    def start(self):
        self.initialize()
        if self.check_delay > 0:
            # The specific case (check_delay < 0)
            # so with no period callback to manage_watchers
            # is probably "unit tests only"
            self.caller = AsyncPeriodicCallback(self.arbiter.manage_watchers,
                                                self.check_delay)
            self.caller.start()
        self.started = True

    def stop(self):
        if self.started:
            if self.caller is not None:
                self.caller.stop()
            try:
                self.stream.flush()
                self.stream.close()
            except (IOError, zmq.ZMQError):
                pass
            self.ctrl_socket.close()
        self.sys_hdl.stop()

    def handle_message(self, raw_msg):
        try:
            # Handle garbage messages,
            # which are not originating from circus
            cid, msg = raw_msg
        except (TypeError, ValueError):
            logger.warning("got unexpected message %s", raw_msg)
            return

        msg = msg.strip()

        if not msg:
            self.send_response(None, cid, msg, "error: empty command")
        else:
            logger.debug("got message %s", msg)
            self.dispatch((cid, msg))

    def handle_autodiscover_message(self, fd_no, type):
        __, address = self.udp_socket.recvfrom(1024)
        self.udp_socket.sendto(json.dumps({'endpoint': self.endpoint}),
                               address)

    def _dispatch_callback_future(self, msg, cid, mid, cast, cmd_name,
                                  send_resp, future):
        exception = check_future_exception_and_log(future)
        if exception is not None:
            if send_resp:
                self.send_error(mid, cid, msg, "server error", cast=cast,
                                errno=errors.BAD_MSG_DATA_ERROR)
        else:
            resp = future.result()
            if send_resp:
                self._dispatch_callback(msg, cid, mid, cast, cmd_name, resp)

    def _dispatch_callback(self, msg, cid, mid, cast, cmd_name, resp=None):
        if resp is None:
            resp = ok()

        if not isinstance(resp, (dict, list)):
            msg = "msg %r tried to send a non-dict: %s" % (msg, str(resp))
            logger.error("msg %r tried to send a non-dict: %s", msg, str(resp))
            return self.send_error(mid, cid, msg, "server error", cast=cast,
                                   errno=errors.BAD_MSG_DATA_ERROR)

        if isinstance(resp, list):
            resp = {"results": resp}

        self.send_ok(mid, cid, msg, resp, cast=cast)

        if cmd_name.lower() == "quit":
            if cid is not None:
                self.stream.flush()

    def dispatch(self, job):
        cid, msg = job
        try:
            json_msg = json.loads(msg)
        except ValueError:
            return self.send_error(None, cid, msg, "json invalid",
                                   errno=errors.INVALID_JSON)

        mid = json_msg.get('id')
        cmd_name = json_msg.get('command')
        properties = json_msg.get('properties', {})
        cast = json_msg.get('msg_type') == "cast"

        try:
            cmd = self.commands[cmd_name.lower()]
        except KeyError:
            error_ = "unknown command: %r" % cmd_name
            return self.send_error(mid, cid, msg, error_, cast=cast,
                                   errno=errors.UNKNOWN_COMMAND)

        try:
            cmd.validate(properties)
            resp = cmd.execute(self.arbiter, properties)
            if isinstance(resp, Future):
                if properties.get('waiting', False):
                    cb = functools.partial(self._dispatch_callback_future, msg,
                                           cid, mid, cast, cmd_name, True)
                    resp.add_done_callback(cb)
                else:
                    cb = functools.partial(self._dispatch_callback_future, msg,
                                           cid, mid, cast, cmd_name, False)
                    resp.add_done_callback(cb)
                    self._dispatch_callback(msg, cid, mid, cast,
                                            cmd_name, None)
            else:
                self._dispatch_callback(msg, cid, mid, cast,
                                        cmd_name, resp)
        except MessageError as e:
            return self.send_error(mid, cid, msg, str(e), cast=cast,
                                   errno=errors.MESSAGE_ERROR)
        except ConflictError as e:
            # conflicts between two commands, sending error...
            return self.send_error(mid, cid, msg, str(e), cast=cast,
                                   errno=errors.COMMAND_ERROR)
        except OSError as e:
            return self.send_error(mid, cid, msg, str(e), cast=cast,
                                   errno=errors.OS_ERROR)
        except:  # noqa: E722
            exctype, value = sys.exc_info()[:2]
            tb = traceback.format_exc()
            reason = "command %r: %s" % (msg, value)
            logger.debug("error: command %r: %s\n\n%s", msg, value, tb)
            return self.send_error(mid, cid, msg, reason, tb, cast=cast,
                                   errno=errors.COMMAND_ERROR)

    def send_error(self, mid, cid, msg, reason="unknown", tb=None, cast=False,
                   errno=errors.NOT_SPECIFIED):
        resp = error(reason=reason, tb=tb, errno=errno)
        self.send_response(mid, cid, msg, resp, cast=cast)

    def send_ok(self, mid, cid, msg, props=None, cast=False):
        resp = ok(props)
        self.send_response(mid, cid, msg, resp, cast=cast)

    def send_response(self, mid, cid, msg, resp, cast=False):
        if cast:
            return

        if cid is None:
            return

        if isinstance(resp, str):
            raise DeprecationWarning('Takes only a mapping')

        resp['id'] = mid
        resp = json.dumps(resp)

        logger.debug("sending response %s", resp)

        try:
            self.stream.send(cid, zmq.SNDMORE)
            self.stream.send(resp)
        except (IOError, zmq.ZMQError) as e:
            logger.debug("Received %r - Could not send back %r - %s", msg,
                         resp, str(e))



from circus.exc import ArgumentError, MessageError
from circus import util
import warnings
try:
    import resource
except ImportError:
    resource = None     # NOQA

_HOOKS = ('before_start', 'after_start', 'before_stop', 'after_stop',
          'before_spawn', 'after_spawn', 'before_signal', 'after_signal',
          'before_reap', 'after_reap', 'extended_stats')


def convert_option(key, val):
    if key == "numprocesses":
        return int(val)
    elif key == "warmup_delay":
        return float(val)
    elif key == "working_dir":
        return val
    elif key == "uid":
        return val
    elif key == "gid":
        return val
    elif key == "send_hup":
        return util.to_bool(val)
    elif key == "stop_signal":
        return util.to_signum(val)
    elif key == "stop_children":
        return util.to_bool(val)
    elif key == "shell":
        return util.to_bool(val)
    elif key == "copy_env":
        return util.to_bool(val)
    elif key == "env":
        return util.parse_env_dict(val)
    elif key == "cmd":
        return val
    elif key == "args":
        return val
    elif key == "retry_in":
        return float(val)
    elif key == "max_retry":
        return int(val)
    elif key == "graceful_timeout":
        return float(val)
    elif key == 'max_age':
        return int(val)
    elif key == 'max_age_variance':
        return int(val)
    elif key == 'respawn':
        return util.to_bool(val)
    elif key == "singleton":
        return util.to_bool(val)
    elif key == "close_child_stdin":
        return util.to_bool(val)
    elif key == "close_child_stdout":
        return util.to_bool(val)
    elif key == "close_child_stderr":
        return util.to_bool(val)
    elif key.startswith('stderr_stream.') or key.startswith('stdout_stream.'):
        subkey = key.split('.', 1)[-1]
        if subkey in ('max_bytes', 'backup_count'):
            return int(val)
        return val
    elif key == 'hooks':
        res = {}
        for hook in val.split(','):
            if hook == '':
                continue
            hook = hook.split(':')
            if len(hook) != 2:
                raise ArgumentError(hook)

            name, value = hook
            if name not in _HOOKS:
                raise ArgumentError(name)

            res[name] = value

        return res
    elif key.startswith('hooks.'):
        # we can also set a single hook
        name = key.split('.', 1)[-1]
        if name not in _HOOKS:
            raise ArgumentError(name)
        return val
    elif key.startswith('rlimit_'):
        return int(val)

    raise ArgumentError("unknown key %r" % key)


def validate_option(key, val):
    valid_keys = ('numprocesses', 'warmup_delay', 'working_dir', 'uid',
                  'gid', 'send_hup', 'stop_signal', 'stop_children',
                  'shell', 'env', 'cmd', 'args', 'copy_env', 'retry_in',
                  'max_retry', 'graceful_timeout', 'stdout_stream',
                  'stderr_stream', 'max_age', 'max_age_variance', 'respawn',
                  'singleton', 'hooks', 'close_child_stdin',
                  'close_child_stdout', 'close_child_stderr')

    valid_prefixes = ('stdout_stream.', 'stderr_stream.', 'hooks.', 'rlimit_')

    def _valid_prefix():
        for prefix in valid_prefixes:
            if key.startswith('%s' % prefix):
                return True
        return False

    if key not in valid_keys and not _valid_prefix():
        raise MessageError('unknown key %r' % key)

    if key in ('numprocesses', 'max_retry', 'max_age', 'max_age_variance',
               'stop_signal'):
        if not isinstance(val, int):
            raise MessageError("%r isn't an integer" % key)

    elif key in ('warmup_delay', 'retry_in', 'graceful_timeout',):
        if not isinstance(val, (int, float)):
            raise MessageError("%r isn't a number" % key)

    elif key in ('uid', 'gid',):
        if not isinstance(val, int) and not isinstance(val, str):
            raise MessageError("%r isn't an integer or string" % key)

    elif key in ('send_hup', 'shell', 'copy_env', 'respawn', 'stop_children',
                 'close_child_stdin', 'close_child_stdout',
                 'close_child_stderr'):
        if not isinstance(val, bool):
            raise MessageError("%r isn't a valid boolean" % key)

    elif key in ('env', ):
        if not isinstance(val, dict):
            raise MessageError("%r isn't a valid object" % key)

        for k, v in val.items():
            if not isinstance(v, str):
                raise MessageError("%r isn't a string" % k)

    elif key == 'hooks':
        if not isinstance(val, dict):
            raise MessageError("%r isn't a valid hook dict" % val)

        for key in val:
            if key not in _HOOKS:
                raise MessageError("Unknown hook %r" % val)

    elif key in ('stderr_stream', 'stdout_stream'):
        if not isinstance(val, dict):
            raise MessageError("%r isn't a valid object" % key)
        if 'class' not in val:
            raise MessageError("%r must have a 'class' key" % key)
        if 'refresh_time' in val:
            warnings.warn("'refresh_time' is deprecated and not useful "
                          "anymore for %r" % key)

    elif key.startswith('rlimit_'):
        if resource:
            rlimit_key = key[7:]
            rlimit_int = getattr(
                resource, 'RLIMIT_' + rlimit_key.upper(), None
            )
            if rlimit_int is None:
                raise MessageError("%r isn't a valid rlimit setting" % key)
        else:
            raise MessageError("rlimit options are not supported on this"
                               " platform")

        # note that a null val means RLIM_INFINITY
        if val is not None and not isinstance(val, int):
            raise MessageError("%r rlimit value isn't a valid int" % val)



from circus import ArbiterHandler as _ArbiterHandler


class ArbiterHandler(_ArbiterHandler):
    def __call__(self, watchers, **kwargs):
        return super(ArbiterHandler, self).__call__(watchers, **kwargs)

    def _get_arbiter_klass(self, background):
        if background:
            raise NotImplementedError
        else:
            from circus.green.arbiter import Arbiter   # NOQA
        return Arbiter


get_arbiter = ArbiterHandler()



import errno
import os
import tempfile
from datetime import datetime
import time as time_
import re
from stat import ST_DEV, ST_INO, ST_MTIME
from circus import logger
from circus.util import to_str


class _FileStreamBase(object):
    """Base class for all file writer handler classes"""
    # You may want to use another now or fromtimestamp method
    # (not naive or a mock).
    now = datetime.now
    fromtimestamp = datetime.fromtimestamp

    def __init__(self, filename, time_format):
        if filename is None:
            fd, filename = tempfile.mkstemp()
            os.close(fd)
        self._filename = filename
        self._file = self._open()
        self._time_format = time_format
        self._buffer = []  # XXX - is this really needed?

    def _open(self):
        return open(self._filename, 'a+')

    def open(self):
        if self._file.closed:
            self._file = self._open()

    def close(self):
        self._file.close()

    def write_data(self, data):
        # data to write on file
        file_data = to_str(data['data'])

        # If we want to prefix the stream with the current datetime
        if self._time_format is not None:
            if 'timestamp' in data:
                time = self.fromtimestamp(data['timestamp'])
            else:
                time = self.now()
            time = time.strftime(self._time_format)
            prefix = '{time} [{pid}] | '.format(time=time, pid=data['pid'])
            file_data = prefix + file_data.rstrip('\n')
            file_data = file_data.replace('\n', '\n' + prefix)
            file_data += '\n'

        # writing into the file
        try:
            self._file.write(file_data)
        except Exception:
            file_data = file_data.encode('latin-1', errors='replace')
            file_data = file_data.decode('latin-1')
            self._file.write(file_data)

        self._file.flush()


class FileStream(_FileStreamBase):
    def __init__(self, filename=None, max_bytes=0, backup_count=0,
                 time_format=None, **kwargs):
        '''
        File writer handler which writes output to a file, allowing rotation
        behaviour based on Python's ``logging.handlers.RotatingFileHandler``.

        By default, the file grows indefinitely. You can specify particular
        values of max_bytes and backup_count to allow the file to rollover at
        a predetermined size.

        Rollover occurs whenever the current log file is nearly max_bytes in
        length. If backup_count is >= 1, the system will successively create
        new files with the same pathname as the base file, but with extensions
        ".1", ".2" etc. appended to it. For example, with a backup_count of 5
        and a base file name of "app.log", you would get "app.log",
        "app.log.1", "app.log.2", ... through to "app.log.5". The file being
        written to is always "app.log" - when it gets filled up, it is closed
        and renamed to "app.log.1", and if files "app.log.1", "app.log.2" etc.
        exist, then they are renamed to "app.log.2", "app.log.3" etc.
        respectively.

        If max_bytes is zero, rollover never occurs.

        You may also configure the timestamp format as defined by
        datetime.strftime.

        Here is an example: ::

          [watcher:foo]
          cmd = python -m myapp.server
          stdout_stream.class = FileStream
          stdout_stream.filename = /var/log/circus/out.log
          stdout_stream.time_format = %Y-%m-%d %H:%M:%S
        '''
        super(FileStream, self).__init__(filename, time_format)
        self._max_bytes = int(max_bytes)
        self._backup_count = int(backup_count)

    def __call__(self, data):
        if self._should_rollover(data['data']):
            self._do_rollover()

        self.write_data(data)

    def _do_rollover(self):
        """
        Do a rollover, as described in __init__().
        """
        if self._file:
            self._file.close()
            self._file = None
        if self._backup_count > 0:
            for i in range(self._backup_count - 1, 0, -1):
                sfn = "%s.%d" % (self._filename, i)
                dfn = "%s.%d" % (self._filename, i + 1)
                if os.path.exists(sfn):
                    logger.debug("Log rotating %s -> %s" % (sfn, dfn))
                    if os.path.exists(dfn):
                        os.remove(dfn)
                    os.rename(sfn, dfn)
            dfn = self._filename + ".1"
            if os.path.exists(dfn):
                os.remove(dfn)
            os.rename(self._filename, dfn)
            logger.debug("Log rotating %s -> %s" % (self._filename, dfn))
        self._file = self._open()

    def _should_rollover(self, raw_data):
        """
        Determine if rollover should occur.

        Basically, see if the supplied raw_data would cause the file to exceed
        the size limit we have.
        """
        if self._file is None:                 # delay was set...
            self._file = self._open()
        if self._max_bytes > 0:                   # are we rolling over?
            self._file.seek(0, 2)  # due to non-posix-compliant Windows feature
            if self._file.tell() + len(raw_data) >= self._max_bytes:
                return 1
        return 0


class WatchedFileStream(_FileStreamBase):
    def __init__(self, filename=None, time_format=None, **kwargs):
        '''
        File writer handler which writes output to a file, allowing an external
        log rotation process to handle rotation, like Python's
        ``logging.handlers.WatchedFileHandler``.

        By default, the file grows indefinitely, and you are responsible for
        ensuring that log rotation happens with some external tool like
        logrotate.

        You may also configure the timestamp format as defined by
        datetime.strftime.

        Here is an example: ::

          [watcher:foo]
          cmd = python -m myapp.server
          stdout_stream.class = WatchedFileStream
          stdout_stream.filename = /var/log/circus/out.log
          stdout_stream.time_format = %Y-%m-%d %H:%M:%S
        '''
        super(WatchedFileStream, self).__init__(filename, time_format)
        self.dev, self.ino = -1, -1
        self._statfile()

    def _statfile(self):
        stb = os.fstat(self._file.fileno())
        self.dev, self.ino = stb[ST_DEV], stb[ST_INO]

    def _statfilename(self):
        try:
            stb = os.stat(self._filename)
            return stb[ST_DEV], stb[ST_INO]
        except OSError as err:
            if err.errno == errno.ENOENT:
                return -1, -1
            else:
                raise

    def __call__(self, data):
        # stat the filename to see if the file we opened still exists. If the
        # ino or dev doesn't match, we need to open a new file handle
        dev, ino = self._statfilename()
        if dev != self.dev or ino != self.ino:
            self._file.flush()
            self._file.close()
            self._file = self._open()
            self._statfile()

        self.write_data(data)


_MIDNIGHT = 24 * 60 * 60  # number of seconds in a day


class TimedRotatingFileStream(FileStream):

    def __init__(self, filename=None, backup_count=0, time_format=None,
                 rotate_when=None, rotate_interval=1, utc=False, **kwargs):
        '''
        File writer handler which writes output to a file, allowing rotation
        behaviour based on Python's
        ``logging.handlers.TimedRotatingFileHandler``.

        The parameters are the same as ``FileStream`` except max_bytes.

        In addition you can specify extra parameters:

        - utc: if True, times in UTC will be used. otherwise local time is
          used. Default: False.
        - rotate_when: the type of interval. Can be S, M, H, D,
          'W0'-'W6' or 'midnight'. See Python's TimedRotatingFileHandler
          for more information.
        - rotate_interval: Rollover interval in seconds. Default: 1

        Here is an example: ::

          [watcher:foo]
          cmd = python -m myapp.server
          stdout_stream.class = TimedRotatingFileStream
          stdout_stream.filename = /var/log/circus/out.log
          stdout_stream.time_format = %Y-%m-%d %H:%M:%S
          stdout_stream.utc = True
          stdout_stream.rotate_when = H
          stdout_stream.rotate_interval = 1

        '''
        super(TimedRotatingFileStream,
              self).__init__(filename=filename, backup_count=backup_count,
                             time_format=time_format, utc=False,
                             **kwargs)

        self._utc = bool(utc)
        self._when = rotate_when

        if self._when == "S":
            self._interval = 1
            self._suffix = "%Y%m%d%H%M%S"
            self._ext_match = r"^\d{4}\d{2}\d{2}\d{2}\d{2}\d{2}$"
        elif self._when == "M":
            self._interval = 60
            self._suffix = "%Y%m%d%H%M"
            self._ext_match = r"^\d{4}\d{2}\d{2}\d{2}\d{2}$"
        elif self._when == "H":
            self._interval = 60 * 60
            self._suffix = "%Y%m%d%H"
            self._ext_match = r"^\d{4}\d{2}\d{2}\d{2}$"
        elif self._when in ("D", "MIDNIGHT"):
            self._interval = 60 * 60 * 24
            self._suffix = "%Y%m%d"
            self._ext_match = r"^\d{4}\d{2}\d{2}$"
        elif self._when.startswith("W"):
            self._interval = 60 * 60 * 24 * 7
            if len(self._when) != 2:
                raise ValueError("You must specify a day for weekly\
rollover from 0 to 6 (0 is Monday): %s" % self._when)
            if self._when[1] < "0" or self._when[1] > "6":
                raise ValueError("Invalid day specified\
for weekly rollover: %s" % self._when)
            self._day_of_week = int(self._when[1])
            self._suffix = "%Y%m%d"
            self._ext_match = r"^\d{4}\d{2}\d{2}$"
        else:
            raise ValueError("Invalid rollover interval specified: %s" %
                             self._when)

        self._ext_match = re.compile(self._ext_match)
        self._interval = self._interval * int(rotate_interval)

        if os.path.exists(self._filename):
            t = os.stat(self._filename)[ST_MTIME]
        else:
            t = int(time_.time())
        self._rollover_at = self._compute_rollover(t)

    def _do_rollover(self):
        if self._file:
            self._file.close()
            self._file = None

        current_time = int(time_.time())
        dst_now = time_.localtime(current_time)[-1]
        t = self._rollover_at - self._interval
        if self._utc:
            time_touple = time_.gmtime(t)
        else:
            time_touple = time_.localtime(t)
            dst_then = time_touple[-1]
            if dst_now != dst_then:
                if dst_now:
                    addend = 3600
                else:
                    addend = -3600
                time_touple = time_.localtime(t + addend)

        dfn = self._filename + "." + time_.strftime(self._suffix, time_touple)

        if os.path.exists(dfn):
            os.remove(dfn)

        if os.path.exists(self._filename):
            os.rename(self._filename, dfn)
            logger.debug("Log rotating %s -> %s" % (self._filename, dfn))

        if self._backup_count > 0:
            for f in self._get_files_to_delete():
                os.remove(f)

        self._file = self._open()

        new_rollover_at = self._compute_rollover(current_time)
        while new_rollover_at <= current_time:
            new_rollover_at = new_rollover_at + self._interval
        self._rollover_at = new_rollover_at

    def _compute_rollover(self, current_time):
        result = current_time + self._interval

        if self._when == "MIDNIGHT" or self._when.startswith("W"):
            if self._utc:
                t = time_.gmtime(current_time)
            else:
                t = time_.localtime(current_time)
            current_hour = t[3]
            current_minute = t[4]
            current_second = t[5]

            r = _MIDNIGHT - ((current_hour * 60 + current_minute) *
                             60 + current_second)
            result = current_time + r

            if self._when.startswith("W"):
                day = t[6]
                if day != self._day_of_week:
                    days_to_wait = self._day_of_week - day
                else:
                    days_to_wait = 6 - day + self._day_of_week + 1
                new_rollover_at = result + (days_to_wait * (60 * 60 * 24))
                if not self._utc:
                    dst_now = t[-1]
                    dst_at_rollover = time_.localtime(new_rollover_at)[-1]
                    if dst_now != dst_at_rollover:
                        if not dst_now:
                            addend = -3600
                        else:
                            addend = 3600
                        new_rollover_at += addend
                result = new_rollover_at

        return result

    def _get_files_to_delete(self):
        dirname, basename = os.path.split(self._filename)
        prefix = basename + "."
        plen = len(prefix)

        result = []
        for filename in os.listdir(dirname):
            if filename[:plen] == prefix:
                suffix = filename[plen:]
                if self._ext_match.match(suffix):
                    result.append(os.path.join(dirname, filename))
        result.sort()
        if len(result) < self._backup_count:
            return []
        return result[:len(result) - self._backup_count]

    def _should_rollover(self, raw_data):
        """
        Determine if rollover should occur.

        record is not used, as we are just comparing times, but it is needed so
        the method signatures are the same
        """
        t = int(time_.time())
        if t >= self._rollover_at:
            return 1
        return 0



from circus.commands.base import Command
import operator


class ListSockets(Command):
    """\
        Get the list of sockets
        =======================

        ZMQ Message
        -----------


        To get the list of sockets::

            {
                "command": "listsockets",
            }


        The response return a list of json mappings with keys for fd, name,
        host and port.

        Command line
        ------------

        ::

            $ circusctl listsockets
    """
    name = "listsockets"

    def message(self, *args, **opts):
        return self.make_message()

    def execute(self, arbiter, props):

        def _get_info(socket):
            sock = {'fd': socket.fileno(),
                    'name': socket.name,
                    'backlog': socket.backlog}

            if socket.host is not None:
                sock['host'] = socket.host
                sock['port'] = socket.port
            else:
                sock['path'] = socket.path

            return sock

        sockets = [_get_info(socket) for socket in arbiter.sockets.values()]
        sockets.sort(key=operator.itemgetter('fd'))
        return {"sockets": sockets}

    def console_msg(self, msg):
        if 'sockets' in msg:
            sockets = []
            for sock in msg['sockets']:
                d = "%(fd)d:socket '%(name)s' "
                if 'path' in sock:
                    d = (d + 'at %(path)s') % sock
                else:
                    d = (d + 'at %(host)s:%(port)d') % sock

                sockets.append(d)

            return "\n".join(sockets)

        return self.console_error(msg)



import sys
import random

from datetime import datetime
from queue import Queue, Empty  # noqa: F401

from circus.util import resolve_name, to_str
from circus.stream.file_stream import FileStream
from circus.stream.file_stream import WatchedFileStream  # noqa: F401
from circus.stream.file_stream import TimedRotatingFileStream  # noqa: F401
from circus.stream.redirector import Redirector  # noqa: F401


class QueueStream(Queue):

    def __init__(self, **kwargs):
        Queue.__init__(self)

    def __call__(self, data):
        self.put(data)

    def close(self):
        pass


class StdoutStream(object):
    def __init__(self, **kwargs):
        pass

    def __call__(self, data):
        sys.stdout.write(to_str(data['data']))
        sys.stdout.flush()

    def close(self):
        pass


class FancyStdoutStream(StdoutStream):
    """
    Write output from watchers using different colors along with a
    timestamp.

    If no color is selected a color will be chosen at random. The
    available ascii colors are:

      - red
      - green
      - yellow
      - blue
      - magenta
      - cyan
      - white

    You may also configure the timestamp format as defined by
    datetime.strftime. The default is: ::

      %Y-%m-%d %H:%M:%S

    Here is an example: ::

      [watcher:foo]
      cmd = python -m myapp.server
      stdout_stream.class = FancyStdoutStream
      stdout_stream.color = green
      stdout_stream.time_format = '%Y/%m/%d | %H:%M:%S'
    """

    # colors in order according to the ascii escape sequences
    colors = ['red', 'green', 'yellow', 'blue',
              'magenta', 'cyan', 'white']

    # Where we write output
    out = sys.stdout

    # Generate a datetime object
    now = datetime.now
    fromtimestamp = datetime.fromtimestamp

    def __init__(self, color=None, time_format=None, **kwargs):
        super(FancyStdoutStream, self).__init__(**kwargs)
        self.time_format = time_format or '%Y-%m-%d %H:%M:%S'
        if color not in self.colors:
            color = random.choice(self.colors)
        self.color_code = self.colors.index(color) + 1

    def prefix(self, data):
        """
        Create a prefix for each line.

        This includes the ansi escape sequence for the color. This
        will not work on windows. For something more robust there is a
        good discussion over on Stack Overflow:

        http://stackoverflow.com/questions/287871
        """
        pid = data['pid']
        if 'timestamp' in data:
            time = self.fromtimestamp(data['timestamp'])
        else:
            time = self.now()
        time = time.strftime(self.time_format)

        # start the coloring with the ansi escape sequence
        color = '\033[0;3%s;40m' % self.color_code

        prefix = '{time} [{pid}] | '.format(pid=pid, time=time)
        return color + prefix

    def __call__(self, data):
        for line in to_str(data['data']).split('\n'):
            if line:
                self.out.write(self.prefix(data))
                self.out.write(line)
                # stop coloring
                self.out.write('\033[0m\n')
                self.out.flush()


def get_stream(conf, reload=False):
    if conf:
        # we can have 'stream' or 'class' or 'filename'
        if 'class' in conf:
            class_name = conf.pop('class')
            if "." not in class_name:
                cls = globals()[class_name]
                inst = cls(**conf)
            else:
                inst = resolve_name(class_name, reload=reload)(**conf)
        elif 'stream' in conf:
            inst = conf['stream']
        elif 'filename' in conf:
            inst = FileStream(**conf)
        else:
            raise ValueError("stream configuration invalid")

        return inst



from circus.consumer import CircusConsumer as _CircusConsumer

from zmq.green import Context, Poller, POLLIN


class CircusConsumer(_CircusConsumer):
    def _init_context(self, context):
        self.context = context or Context()

    def _init_poller(self):
        self.poller = Poller()
        self.poller.register(self.pubsub_socket, POLLIN)



from circus.commands.base import Command
from circus.exc import ArgumentError
from circus.util import TransformableFuture


class Reload(Command):
    """\
        Reload the arbiter or a watcher
        ===============================

        This command reloads all the process in a watcher or all watchers. This
        will happen in one of 3 ways:

        * If graceful is false, a simple restart occurs.
        * If `send_hup` is true for the watcher, a HUP signal is sent to each
          process.
        * Otherwise:
            * If sequential is false, the arbiter will attempt to spawn
              `numprocesses` new processes. If the new processes are spawned
              successfully, the result is that all of the old processes are
              stopped, since by default the oldest processes are stopped when
              the actual number of processes for a watcher is greater than
              `numprocesses`.
            * If sequential is true, the arbiter will restart each process
              in a sequential way (with a `warmup_delay` pause between each
              step)


        ZMQ Message
        -----------

        ::

            {
                "command": "reload",
                "properties": {
                    "name": '<name>",
                    "graceful": true,
                    "sequential": false,
                    "waiting": False
                }
            }

        The response return the status "ok". If the property graceful is
        set to true the processes will be exited gracefully.

        If the property name is present, then the reload will be applied
        to the watcher.


        Command line
        ------------

        ::

            $ circusctl reload [<name>] [--terminate] [--waiting]
                                        [--sequential]

        Options
        +++++++

        - <name>: name of the watcher
        - --terminate; quit the node immediately

    """
    name = "reload"
    options = (Command.options + Command.waiting_options +
               [('', 'sequential', False, "sequential reload")] +
               [('', 'terminate', False, "stop immediately")])

    def message(self, *args, **opts):
        if len(args) > 1:
            raise ArgumentError("invalid number of arguments")

        graceful = not opts.get("terminate", False)
        waiting = opts.get("waiting", False)
        sequential = opts.get("sequential", False)
        if len(args) == 1:
            return self.make_message(name=args[0], graceful=graceful,
                                     waiting=waiting, sequential=sequential)
        else:
            return self.make_message(graceful=graceful, waiting=waiting,
                                     sequential=sequential)

    def execute(self, arbiter, props):
        graceful = props.get('graceful', True)
        sequential = props.get('sequential', False)
        if 'name' in props:
            watcher = self._get_watcher(arbiter, props['name'])
            if props.get('waiting'):
                resp = TransformableFuture()
                resp.set_upstream_future(watcher.reload(
                    graceful=graceful,
                    sequential=sequential))
                resp.set_transform_function(lambda x: {"info": x})
                return resp
            return watcher.reload(graceful=graceful,
                                  sequential=sequential)
        else:
            return arbiter.reload(graceful=graceful, sequential=sequential)




from circus.plugins.statsd import BaseObserver

try:
    import redis
except ImportError:
    raise ImportError("This plugin requires the redis-lib to run.")


class RedisObserver(BaseObserver):

    name = 'redis_observer'
    default_app_name = "redis_observer"

    OBSERVE = ['pubsub_channels', 'connected_slaves', 'lru_clock',
               'connected_clients', 'keyspace_misses', 'used_memory',
               'used_memory_peak', 'total_commands_processed',
               'used_memory_rss', 'total_connections_received',
               'pubsub_patterns', 'used_cpu_sys', 'used_cpu_sys_children',
               'blocked_clients', 'used_cpu_user', 'client_biggest_input_buf',
               'mem_fragmentation_ratio', 'expired_keys', 'evicted_keys',
               'client_longest_output_list', 'uptime_in_seconds',
               'keyspace_hits']

    def __init__(self, *args, **config):
        super(RedisObserver, self).__init__(*args, **config)
        self.redis = redis.from_url(config.get("redis_url",
                                    "redis://localhost:6379/0"),
                                    float(config.get("timeout", 5)))

        self.restart_on_timeout = config.get("restart_on_timeout", None)

    def look_after(self):
        try:
            info = self.redis.info()
        except redis.ConnectionError:
            self.statsd.increment("redis_stats.error")
            if self.restart_on_timeout:
                self.cast("restart", name=self.restart_on_timeout)
                self.statsd.increment("redis_stats.restart_on_error")
            return

        for key in self.OBSERVE:
            self.statsd.gauge("redis_stats.%s" % key, info[key])



from circus.client import CircusClient as _CircusClient

from zmq.green import Context, Poller, POLLIN


class CircusClient(_CircusClient):
    def _init_context(self, context):
        self.context = context or Context.instance()

    def _init_poller(self):
        self.poller = Poller()
        self.poller.register(self.socket, POLLIN)



from circus.commands.base import Command
from circus.exc import ArgumentError


class Status(Command):
    """\
        Get the status of a watcher or all watchers
        ===========================================

        This command start get the status of a watcher or all watchers.

        ZMQ Message
        -----------

        ::

            {
                "command": "status",
                "properties": {
                    "name": '<name>",
                }
            }

        The response return the status "active" or "stopped" or the
        status / watchers.


        Command line
        ------------

        ::

            $ circusctl status [<name>]

        Options
        +++++++

        - <name>: name of the watcher

        Example
        +++++++

        ::

            $ circusctl status dummy
            active
            $ circusctl status
            dummy: active
            dummy2: active
            refuge: active

    """

    name = "status"

    def message(self, *args, **opts):
        if len(args) > 1:
            raise ArgumentError("message invalid")

        if len(args) == 1:
            return self.make_message(name=args[0])
        else:
            return self.make_message()

    def execute(self, arbiter, props):
        if 'name' in props:
            watcher = self._get_watcher(arbiter, props['name'])
            return {"status": watcher.status()}
        else:
            return {"statuses": arbiter.statuses()}

    def console_msg(self, msg):
        if "statuses" in msg:
            statuses = msg.get("statuses")
            watchers = sorted(statuses)
            return "\n".join(["%s: %s" % (watcher, statuses[watcher])
                              for watcher in watchers])
        elif "status" in msg and "status" != "error":
            return msg.get("status")
        return self.console_error(msg)



from circus.commands.base import Command
from circus.exc import ArgumentError, MessageError
from circus.util import convert_opt


class Get(Command):
    """\
        Get the value of specific watcher options
        =========================================

        This command can be used to query the current value of one or
        more watcher options.

        ZMQ Message
        -----------

        ::

            {
                "command": "get",
                "properties": {
                    "keys": ["key1, "key2"]
                    "name": "nameofwatcher"
                }
            }

        A request message contains two properties:

        - keys: list, The option keys for which you want to get the values
        - name: name of watcher

        The response object has a property ``options`` which is a
        dictionary of option names and values.

        eg::

            {
                "status": "ok",
                "options": {
                    "graceful_timeout": 300,
                    "send_hup": True,
                },
                time': 1332202594.754644
            }


        Command line
        ------------

        ::

            $ circusctl get <name> <key1> <key2>

    """

    name = "get"
    properties = ['name', 'keys']

    def message(self, *args, **opts):
        if len(args) < 2:
            raise ArgumentError("Invalid number of arguments")

        return self.make_message(name=args[0], keys=args[1:])

    def execute(self, arbiter, props):
        watcher = self._get_watcher(arbiter, props.get('name'))

        # get options values. It return an error if one of the asked
        # options isn't found
        options = {}
        for name in props.get('keys', []):
            if name in watcher.optnames:
                options[name] = getattr(watcher, name)
            else:
                raise MessageError("%r option not found" % name)

        return {"options": options}

    def console_msg(self, msg):
        if msg['status'] == "ok":
            ret = []
            for k, v in msg.get('options', {}).items():
                ret.append("%s: %s" % (k, convert_opt(k, v)))
            return "\n".join(ret)
        return self.console_error(msg)



import errno
import os
import stat
import tempfile


class Pidfile(object):
    """
    Manage a PID file. If a specific name is provided
    it and '"%s.oldpid" % name' will be used. Otherwise
    we create a temp file using tempfile.mkstemp.
    """

    def __init__(self, fname):
        self.fname = fname
        self.pid = os.getpid()
        # set permissions to -rw-r--r--
        self.perm_mode = (stat.S_IRUSR | stat.S_IWUSR |
                          stat.S_IRGRP |
                          stat.S_IROTH)

    def create(self, pid):
        pid = int(pid)
        oldpid = self.validate()
        if oldpid:
            if oldpid == pid:
                return
            raise RuntimeError("pid file '{0}' is stale, current pid {1}"
                               .format(self.fname, pid))

        self.pid = pid

        # Write pidfile
        if self.fname:
            fdir = os.path.dirname(self.fname)
            if fdir and not os.path.isdir(fdir):
                raise RuntimeError("{0} doesn't exist. Can't create pidfile"
                                   .format(fdir))
            flags = os.O_CREAT | os.O_WRONLY | os.O_TRUNC
            fd = os.open(self.fname, flags, self.perm_mode)
        else:
            fd, self.fname = tempfile.mkstemp()

        os.chmod(self.fname, self.perm_mode)
        os.write(fd, "{0}\n".format(self.pid).encode('utf-8'))
        os.fsync(fd)
        os.close(fd)

    def rename(self, path):
        self.unlink()
        self.fname = path
        self.create(self.pid)

    def unlink(self):
        """ delete pidfile"""
        try:
            with open(self.fname, "r") as f:
                try:
                    pid1 = int(f.read() or 0)
                except ValueError:
                    pid1 = self.pid

            if pid1 == self.pid:
                os.unlink(self.fname)
        except:  # noqa: E722
            pass

    def validate(self):
        """ Validate pidfile and make it stale if needed"""
        if not self.fname:
            return
        try:
            with open(self.fname, "r") as f:
                try:
                    wpid = int(f.read() or 0)
                except ValueError:
                    return

                if wpid <= 0:
                    return

                try:
                    os.kill(wpid, 0)
                    return wpid
                except OSError as e:
                    if e.args[0] == errno.ESRCH:
                        return
                    raise
        except ValueError:
            return
        except IOError as e:
            if e.args[0] == errno.ENOENT:
                return
            raise



import gevent

from circus.sighandler import SysHandler as _SysHandler


class SysHandler(_SysHandler):

    def _register(self):
        for sig in self.SIGNALS:
            gevent.signal(sig, self.signal, sig)



""" Base class to create Circus subscribers plugins.
"""
import sys
import errno
import uuid
import argparse

import zmq
import zmq.utils.jsonapi as json
from tornado import ioloop
from zmq.eventloop import zmqstream

from circus import logger, __version__
from circus.client import make_message, cast_message
from circus.util import (debuglog, to_bool, resolve_name, configure_logger,
                         DEFAULT_ENDPOINT_DEALER, DEFAULT_ENDPOINT_SUB,
                         get_connection, to_str, to_bytes)


class CircusPlugin(object):
    """Base class to write plugins.

    Options:

    - **context** -- the ZMQ context to use
    - **endpoint** -- the circusd ZMQ endpoint
    - **pubsub_endpoint** -- the circusd ZMQ pub/sub endpoint
    - **check_delay** -- the configured check delay
    - **config** -- free config mapping
    """
    name = ''

    def __init__(self, endpoint, pubsub_endpoint, check_delay, ssh_server=None,
                 **config):
        self.daemon = True
        self.config = config
        self.active = to_bool(config.get('active', True))
        self.pubsub_endpoint = pubsub_endpoint
        self.endpoint = endpoint
        self.check_delay = check_delay
        self.ssh_server = ssh_server
        self._id = to_bytes(uuid.uuid4().hex)
        self.running = False
        self.loop = ioloop.IOLoop()

    @debuglog
    def initialize(self):
        self.context = zmq.Context()
        self.client = self.context.socket(zmq.DEALER)
        self.client.setsockopt(zmq.IDENTITY, self._id)
        get_connection(self.client, self.endpoint, self.ssh_server)
        self.client.linger = 0
        self.sub_socket = self.context.socket(zmq.SUB)
        self.sub_socket.setsockopt(zmq.SUBSCRIBE, b'watcher.')
        self.sub_socket.connect(self.pubsub_endpoint)
        self.substream = zmqstream.ZMQStream(self.sub_socket, self.loop)
        self.substream.on_recv(self.handle_recv)

    @debuglog
    def start(self):
        self.loop.make_current()
        if not self.active:
            raise ValueError('Will not start an inactive plugin')
        self.handle_init()
        self.initialize()
        self.running = True

        while True:
            try:
                self.loop.start()
            except zmq.ZMQError as e:
                logger.debug(str(e))

                if e.errno == errno.EINTR:
                    continue
                elif e.errno == zmq.ETERM:
                    break
                else:
                    logger.debug("got an unexpected error %s (%s)", str(e),
                                 e.errno)
                    raise
            else:
                break

        self.substream.close()
        self.client.close()
        self.sub_socket.close()
        self.context.destroy()

    @debuglog
    def stop(self):
        if not self.running:
            self.loop.close()
            return

        try:
            self.handle_stop()
        finally:
            self.loop.stop()

        self.running = False

    def call(self, command, **props):
        """Sends the command to **circusd**

        Options:

        - **command** -- the command to call
        - **props** -- keyword arguments to add to the call

        Returns the JSON mapping sent back by **circusd**
        """
        msg = make_message(command, **props)
        self.client.send(json.dumps(msg))
        msg = self.client.recv()
        return json.loads(msg)

    def cast(self, command, **props):
        """Fire-and-forget a command to **circusd**

        Options:

        - **command** -- the command to call
        - **props** -- keyword arguments to add to the call
        """
        msg = cast_message(command, **props)
        self.client.send(json.dumps(msg))

    #
    # methods to override.
    #
    def handle_recv(self, data):
        """Receives every event published by **circusd**

        Options:

        - **data** -- a tuple containing the topic and the message.
        """
        raise NotImplementedError()

    def handle_stop(self):
        """Called right before the plugin is stopped by Circus.
        """
        pass

    def handle_init(self):
        """Called right before a plugin is started - in the thread context.
        """
        pass

    @staticmethod
    def split_data(data):
        topic, msg = data
        topic_parts = to_str(topic).split(".")
        return topic_parts[1], topic_parts[2], msg

    @staticmethod
    def load_message(msg):
        return json.loads(msg)


def _cfg2str(cfg):
    return ':::'.join([
         '%s:%s' % (key, val) for key, val in sorted(cfg.items())
    ])


def _str2cfg(data):
    cfg = {}
    if data is None:
        return cfg

    for item in data.split(':::'):
        item = item.split(':', 1)
        if len(item) != 2:
            continue
        key, value = item
        cfg[key.strip()] = value.strip()

    return cfg


def get_plugin_cmd(config, endpoint, pubsub, check_delay, ssh_server,
                   debug=False, loglevel=None, logoutput=None):
    fqn = config['use']
    # makes sure the name exists
    resolve_name(fqn)

    # we're good, serializing the config
    del config['use']
    config = _cfg2str(config)
    cmd = "%s -c 'from circus import plugins;plugins.main()'" % sys.executable
    cmd += ' --endpoint %s' % endpoint
    cmd += ' --pubsub %s' % pubsub
    if ssh_server is not None:
        cmd += ' --ssh %s' % ssh_server
    if len(config) > 0:
        cmd += ' --config %s' % config
    if debug:
        cmd += ' --log-level DEBUG'
    elif loglevel:
        cmd += ' --log-level ' + loglevel
    if logoutput:
        cmd += ' --log-output ' + logoutput
    cmd += ' %s' % fqn
    return cmd


def main():
    parser = argparse.ArgumentParser(description='Runs a plugin.')

    parser.add_argument('--endpoint',
                        help='The circusd ZeroMQ socket to connect to',
                        default=DEFAULT_ENDPOINT_DEALER)

    parser.add_argument('--pubsub',
                        help='The circusd ZeroMQ pub/sub socket to connect to',
                        default=DEFAULT_ENDPOINT_SUB)

    parser.add_argument('--config', help='The plugin configuration',
                        default=None)

    parser.add_argument('--version', action='store_true', default=False,
                        help='Displays Circus version and exits.')

    parser.add_argument('--check-delay', type=float, default=5.,
                        help='Checck delay.')

    parser.add_argument('plugin',
                        help='Fully qualified name of the plugin class.',
                        nargs='?')

    parser.add_argument('--log-level', dest='loglevel', default='info',
                        help="log level")

    parser.add_argument('--log-output', dest='logoutput', default='-',
                        help="log output")

    parser.add_argument('--ssh', default=None, help='SSH Server')

    args = parser.parse_args()

    if args.version:
        print(__version__)
        sys.exit(0)

    if args.plugin is None:
        parser.print_usage()
        sys.exit(0)

    factory = resolve_name(args.plugin)

    # configure the logger
    configure_logger(logger, args.loglevel, args.logoutput, name=factory.name)

    # load the plugin and run it.
    logger.info('Loading the plugin...')
    logger.info('Endpoint: %r' % args.endpoint)
    logger.info('Pub/sub: %r' % args.pubsub)
    plugin = factory(args.endpoint, args.pubsub,
                     args.check_delay, args.ssh,
                     **_str2cfg(args.config))
    logger.info('Starting')
    try:
        plugin.start()
    except KeyboardInterrupt:
        pass
    finally:
        logger.info('Stopping')
        plugin.stop()
    sys.exit(0)


if __name__ == '__main__':
    main()



import socket

from circus.plugins import CircusPlugin
from circus.util import human2bytes, AsyncPeriodicCallback


class StatsdClient(object):

    def __init__(self, host=None, port=None, prefix=None, sample_rate=1):
        self.host = host
        self.port = port
        self.prefix = prefix
        self.sample_rate = sample_rate
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    def send(self, bucket, value, sample_rate=None):
        sample_rate = sample_rate or self.sample_rate
        if sample_rate != 1:
            value += "|@%s" % sample_rate

        if self.prefix:
            bucket = "%s.%s" % (self.prefix, bucket)

        msg = "%s:%s" % (bucket, value)
        self.socket.sendto(msg.encode('utf-8'), (self.host, self.port))

    def decrement(self, bucket, delta=1):
        if delta > 0:
            delta = - delta
        self.increment(bucket, delta)

    def increment(self, bucket, delta=1):
        self.send(bucket, "%d|c" % delta)

    def gauge(self, bucket, value):
        self.send(bucket, "%s|g" % value)

    def timed(self, bucket, value):
        self.send(bucket, "%s|ms" % value)

    def stop(self):
        self.socket.close()


class StatsdEmitter(CircusPlugin):
    """Plugin that sends stuff to statsd
    """
    name = 'statsd'
    default_app_name = "app"

    def __init__(self, endpoint, pubsub_endpoint, check_delay, ssh_server,
                 **config):
        super(StatsdEmitter, self).__init__(endpoint, pubsub_endpoint,
                                            check_delay, ssh_server=ssh_server)
        self.app = config.get('application_name', self.default_app_name)
        self.prefix = 'circus.%s.watcher' % self.app

        # initialize statsd
        self.statsd = StatsdClient(host=config.get('host', 'localhost'),
                                   port=int(config.get('port', '8125')),
                                   prefix=self.prefix,
                                   sample_rate=float(
                                       config.get('sample_rate', '1.0')))

    def handle_recv(self, data):
        watcher_name, action, msg = self.split_data(data)
        self.statsd.increment('%s.%s' % (watcher_name, action))

    def stop(self):
        self.statsd.stop()
        super(StatsdEmitter, self).stop()


class BaseObserver(StatsdEmitter):

    def __init__(self, *args, **config):
        super(BaseObserver, self).__init__(*args, **config)
        self.loop_rate = float(config.get("loop_rate", 60))  # in seconds

    def handle_init(self):
        self.period = AsyncPeriodicCallback(self.look_after,
                                            self.loop_rate * 1000)
        self.period.start()

    def handle_stop(self):
        self.period.stop()
        self.statsd.stop()

    def handle_recv(self, data):
        pass

    def look_after(self):
        raise NotImplementedError()


class FullStats(BaseObserver):

    name = 'full_stats'

    def look_after(self):
        info = self.call("stats")
        if info["status"] == "error":
            self.statsd.increment("_stats.error")
            return

        for name, stats in info['infos'].items():
            if name.startswith("plugin:"):
                # ignore plugins
                continue

            cpus = []
            mems = []
            mem_infos = []

            for sub_name, sub_info in stats.items():
                if isinstance(sub_info, dict):
                    cpus.append(sub_info['cpu'])
                    mems.append(sub_info['mem'])
                    mem_infos.append(human2bytes(sub_info['mem_info1']))
                elif sub_name == "spawn_count":
                    # spawn_count info is in the same level as processes
                    # dict infos, so if spawn_count is given, take it and
                    # continue
                    self.statsd.gauge("_stats.%s.spawn_count" % name,
                                      sub_info)

            self.statsd.gauge("_stats.%s.watchers_num" % name, len(cpus))

            if not cpus:
                # if there are only dead processes, we have an empty list
                # and we can't measure it
                continue

            self.statsd.gauge("_stats.%s.cpu_max" % name, max(cpus))
            self.statsd.gauge("_stats.%s.cpu_sum" % name, sum(cpus))
            self.statsd.gauge("_stats.%s.mem_pct_max" % name, max(mems))
            self.statsd.gauge("_stats.%s.mem_pct_sum" % name, sum(mems))
            self.statsd.gauge("_stats.%s.mem_max" % name, max(mem_infos))
            self.statsd.gauge("_stats.%s.mem_sum" % name, sum(mem_infos))



import signal
import warnings
from circus.plugins.statsd import BaseObserver
from circus.util import to_bool
from circus.util import human2bytes


class ResourceWatcher(BaseObserver):

    def __init__(self, *args, **config):
        super(ResourceWatcher, self).__init__(*args, **config)
        self.watcher = config.get("watcher", None)
        self.service = config.get("service", None)
        if self.service is not None:
            warnings.warn("ResourceWatcher.service is deprecated "
                          "please use ResourceWatcher.watcher instead.",
                          category=DeprecationWarning)
            if self.watcher is None:
                self.watcher = self.service
        if self.watcher is None:
            self.statsd.stop()
            self.loop.close()
            raise NotImplementedError('watcher is mandatory for now.')

        self.max_cpu = float(config.get("max_cpu", 90))     # in %
        self.max_mem = config.get("max_mem")

        if self.max_mem is None:
            self.max_mem = 90.
            self._max_percent = True
        else:
            try:
                self.max_mem = float(self.max_mem)          # float -> %
                self._max_percent = True
            except ValueError:
                self.max_mem = human2bytes(self.max_mem)    # int -> absolute
                self._max_percent = False

        self.min_cpu = config.get("min_cpu")
        if self.min_cpu is not None:
            self.min_cpu = float(self.min_cpu)              # in %
        self.min_mem = config.get("min_mem")
        if self.min_mem is not None:
            try:
                self.min_mem = float(self.min_mem)          # float -> %
                self._min_percent = True
            except ValueError:
                self.min_mem = human2bytes(self.min_mem)    # int -> absolute
                self._min_percent = True
        self.health_threshold = float(config.get("health_threshold",
                                      75))  # in %
        self.max_count = int(config.get("max_count", 3))

        self.process_children = to_bool(config.get("process_children", '0'))
        self.child_signal = int(config.get("child_signal", signal.SIGTERM))

        self._count_over_cpu = {}
        self._count_over_mem = {}
        self._count_under_cpu = {}
        self._count_under_mem = {}
        self._count_health = {}

    def look_after(self):
        info = self.call("stats", name=self.watcher)

        if info["status"] == "error":
            self.statsd.increment("_resource_watcher.%s.error" % self.watcher)
            return

        stats = info['info']

        self._process_index('parent', self._collect_data(stats))
        if not self.process_children:
            return

        for sub_info in stats.values():
            if isinstance(sub_info, dict):
                for child_info in sub_info['children']:
                    data = self._collect_data({child_info['pid']: child_info})
                    self._process_index(child_info['pid'], data)

    def _collect_data(self, stats):
        data = {}
        cpus = []
        mems = []
        mems_abs = []

        for sub_info in stats.values():
            if isinstance(sub_info, dict):
                cpus.append(100 if sub_info['cpu'] == 'N/A' else
                            float(sub_info['cpu']))
                mems.append(100 if sub_info['mem'] == 'N/A' else
                            float(sub_info['mem']))
                mems_abs.append(0 if sub_info['mem_info1'] == 'N/A' else
                                human2bytes(sub_info['mem_info1']))

        if cpus:
            data['max_cpu'] = max(cpus)
            data['max_mem'] = max(mems)
            data['max_mem_abs'] = max(mems_abs)
            data['min_cpu'] = min(cpus)
            data['min_mem'] = min(mems)
            data['min_mem_abs'] = min(mems_abs)
        else:
            # we dont' have any process running. max = 0 then
            data['max_cpu'] = 0
            data['max_mem'] = 0
            data['min_cpu'] = 0
            data['min_mem'] = 0
            data['max_mem_abs'] = 0
            data['min_mem_abs'] = 0

        return data

    def _process_index(self, index, stats):

        if (index not in self._count_over_cpu or
                index not in self._count_over_mem or
                index not in self._count_under_cpu or
                index not in self._count_under_mem or
                index not in self._count_health):
            self._reset_index(index)

        if self.max_cpu and stats['max_cpu'] > self.max_cpu:
            self.statsd.increment("_resource_watcher.%s.over_cpu" %
                                  self.watcher)
            self._count_over_cpu[index] += 1
        else:
            self._count_over_cpu[index] = 0

        if self.min_cpu is not None and stats['min_cpu'] <= self.min_cpu:
            self.statsd.increment("_resource_watcher.%s.under_cpu" %
                                  self.watcher)
            self._count_under_cpu[index] += 1
        else:
            self._count_under_cpu[index] = 0

        if self.max_mem is not None:
            over_percent = (self._max_percent and
                            stats['max_mem'] > self.max_mem)
            over_value = (not self._max_percent and
                          stats['max_mem_abs'] > self.max_mem)

            if over_percent or over_value:
                self.statsd.increment("_resource_watcher.%s.over_memory" %
                                      self.watcher)
                self._count_over_mem[index] += 1
            else:
                self._count_over_mem[index] = 0
        else:
            self._count_over_mem[index] = 0

        if self.min_mem is not None:
            under_percent = (self._min_percent and
                             stats['min_mem'] < self.min_mem)
            under_value = (not self._min_percent and
                           stats['min_mem_abs'] < self.min_mem)

            if under_percent or under_value:
                self.statsd.increment("_resource_watcher.%s.under_memory" %
                                      self.watcher)
                self._count_under_mem[index] += 1
            else:
                self._count_under_mem[index] = 0
        else:
            self._count_under_mem[index] = 0

        max_cpu = stats['max_cpu']
        max_mem = stats['max_mem']

        if (self.health_threshold and
                (max_cpu + max_mem) / 2.0 > self.health_threshold):
            self.statsd.increment("_resource_watcher.%s.over_health" %
                                  self.watcher)
            self._count_health[index] += 1
        else:
            self._count_health[index] = 0

        if max([self._count_over_cpu[index], self._count_under_cpu[index],
                self._count_over_mem[index], self._count_under_mem[index],
                self._count_health[index]]) > self.max_count:
            self.statsd.increment("_resource_watcher.%s.restarting" %
                                  self.watcher)

            # todo: restart only process instead of the whole watcher
            if index == 'parent':
                self.cast("restart", name=self.watcher)
                self._reset_index(index)
            else:
                self.cast(
                    "signal",
                    name=self.watcher,
                    signum=self.child_signal,
                    child_pid=index
                )
                self._remove_index(index)

            self._reset_index(index)

    def _reset_index(self, index):
        self._count_over_cpu[index] = 0
        self._count_over_mem[index] = 0
        self._count_under_cpu[index] = 0
        self._count_under_mem[index] = 0
        self._count_health[index] = 0

    def _remove_index(self, index):
        del self._count_over_cpu[index]
        del self._count_over_mem[index]
        del self._count_under_cpu[index]
        del self._count_under_mem[index]
        del self._count_health[index]

    def stop(self):
        self.statsd.stop()
        super(ResourceWatcher, self).stop()
